"""
Tests for SB02OX - Select stable eigenvalues for discrete-time Riccati (generalized).

This is a callback function for DGGES eigenvalue selection. Returns True for
generalized eigenvalues with modulus less than 1 (stable in discrete-time).

For generalized eigenvalue lambda = (ALPHAR + i*ALPHAI) / BETA:
  stable when |lambda| = sqrt(ALPHAR^2 + ALPHAI^2) / |BETA| < 1
  i.e., when DLAPY2(ALPHAR, ALPHAI) < |BETA|
"""

import numpy as np


def test_sb02ox_stable_real_eigenvalue():
    """
    Test stable real eigenvalue: |lambda| = 0.5 < 1.

    ALPHAR = 1, ALPHAI = 0, BETA = 2 -> |lambda| = 1/2 = 0.5 < 1 -> stable
    """
    from slicot import sb02ox

    result = sb02ox(1.0, 0.0, 2.0)
    assert result is True


def test_sb02ox_stable_real_negative_beta():
    """
    Test stable real eigenvalue with negative BETA.

    ALPHAR = 1, ALPHAI = 0, BETA = -2 -> |lambda| = 1/2 = 0.5 < 1 -> stable
    """
    from slicot import sb02ox

    result = sb02ox(1.0, 0.0, -2.0)
    assert result is True


def test_sb02ox_unstable_real_eigenvalue():
    """
    Test unstable real eigenvalue: |lambda| = 2 > 1.

    ALPHAR = 2, ALPHAI = 0, BETA = 1 -> |lambda| = 2/1 = 2 > 1 -> unstable
    """
    from slicot import sb02ox

    result = sb02ox(2.0, 0.0, 1.0)
    assert result is False


def test_sb02ox_marginal_eigenvalue():
    """
    Test marginally stable eigenvalue: |lambda| = 1.

    ALPHAR = 1, ALPHAI = 0, BETA = 1 -> |lambda| = 1/1 = 1 -> not strictly stable
    """
    from slicot import sb02ox

    result = sb02ox(1.0, 0.0, 1.0)
    assert result is False


def test_sb02ox_stable_complex_eigenvalue():
    """
    Test stable complex eigenvalue.

    ALPHAR = 0.3, ALPHAI = 0.4, BETA = 1 -> |lambda| = 0.5 < 1 -> stable
    """
    from slicot import sb02ox

    result = sb02ox(0.3, 0.4, 1.0)
    assert result is True


def test_sb02ox_unstable_complex_eigenvalue():
    """
    Test unstable complex eigenvalue.

    ALPHAR = 3, ALPHAI = 4, BETA = 1 -> |lambda| = 5 > 1 -> unstable
    """
    from slicot import sb02ox

    result = sb02ox(3.0, 4.0, 1.0)
    assert result is False


def test_sb02ox_zero_eigenvalue():
    """
    Test zero eigenvalue (origin): |lambda| = 0 < 1 -> stable.

    ALPHAR = 0, ALPHAI = 0, BETA = 1 -> |lambda| = 0 < 1 -> stable
    """
    from slicot import sb02ox

    result = sb02ox(0.0, 0.0, 1.0)
    assert result is True


def test_sb02ox_purely_imaginary_stable():
    """
    Test stable purely imaginary eigenvalue.

    ALPHAR = 0, ALPHAI = 0.5, BETA = 1 -> |lambda| = 0.5 < 1 -> stable
    """
    from slicot import sb02ox

    result = sb02ox(0.0, 0.5, 1.0)
    assert result is True


def test_sb02ox_purely_imaginary_unstable():
    """
    Test unstable purely imaginary eigenvalue.

    ALPHAR = 0, ALPHAI = 2, BETA = 1 -> |lambda| = 2 > 1 -> unstable
    """
    from slicot import sb02ox

    result = sb02ox(0.0, 2.0, 1.0)
    assert result is False


def test_sb02ox_mathematical_property():
    """
    Validate mathematical property: stable iff DLAPY2(ALPHAR, ALPHAI) < |BETA|.

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    from slicot import sb02ox

    for _ in range(20):
        alphar = np.random.uniform(-10, 10)
        alphai = np.random.uniform(-10, 10)
        beta = np.random.uniform(-10, 10)

        if abs(beta) < 0.01:
            continue

        result = sb02ox(alphar, alphai, beta)

        modulus = np.hypot(alphar, alphai)
        expected = modulus < abs(beta)

        assert result == expected, \
            f"sb02ox({alphar}, {alphai}, {beta}) = {result}, expected {expected}"


def test_sb02ox_sb02ov_complement():
    """
    Validate sb02ox and sb02ov are complements.

    For eigenvalues away from the unit circle (|lambda| != 1):
    - sb02ox returns True for stable (|lambda| < 1)
    - sb02ov returns True for unstable (|lambda| >= 1)

    For strictly inside or outside the unit circle, they should be opposite.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    from slicot import sb02ox, sb02ov

    for _ in range(20):
        alphar = np.random.uniform(-10, 10)
        alphai = np.random.uniform(-10, 10)
        beta = np.random.uniform(-10, 10)

        if abs(beta) < 0.01:
            continue

        modulus = np.hypot(alphar, alphai)
        if abs(modulus - abs(beta)) < 0.01:
            continue

        stable = sb02ox(alphar, alphai, beta)
        unstable = sb02ov(alphar, alphai, beta)

        assert stable != unstable, \
            f"For alphar={alphar}, alphai={alphai}, beta={beta}: " \
            f"stable={stable}, unstable={unstable} should be opposite"


def test_sb02ox_boundary_case_just_inside():
    """
    Test eigenvalue just inside unit circle.

    |lambda| = 0.999 < 1 -> stable
    """
    from slicot import sb02ox

    result = sb02ox(0.999, 0.0, 1.0)
    assert result is True


def test_sb02ox_boundary_case_just_outside():
    """
    Test eigenvalue just outside unit circle.

    |lambda| = 1.001 > 1 -> unstable
    """
    from slicot import sb02ox

    result = sb02ox(1.001, 0.0, 1.0)
    assert result is False



"""
Tests for MA02MZ - Compute norms of a complex skew-Hermitian matrix.

MA02MZ computes the value of the one norm, or the Frobenius norm, or
the infinity norm, or the element of largest absolute value
of a complex skew-Hermitian matrix.

Note that for skew-Hermitian matrices:
- Diagonal elements are pure imaginary (real part = 0)
- a[j,i] = -conj(a[i,j]) for off-diagonal elements
- The infinity norm equals the one norm
"""
import numpy as np
import pytest
from slicot import ma02mz


def test_ma02mz_max_norm_upper():
    """
    Test max norm (M) with upper triangular storage.

    For a skew-Hermitian matrix, max norm is max(|a_ij|).
    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n = 4

    a = np.zeros((n, n), order='F', dtype=complex)
    for j in range(1, n):
        for i in range(j):
            a[i, j] = np.random.randn() + 1j * np.random.randn()
            a[j, i] = -np.conj(a[i, j])
    for i in range(n):
        a[i, i] = 1j * np.random.randn()

    expected = np.max(np.abs(a))

    norm = ma02mz('M', 'U', a)

    np.testing.assert_allclose(norm, expected, rtol=1e-14)


def test_ma02mz_max_norm_lower():
    """
    Test max norm (M) with lower triangular storage.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n = 5

    a = np.zeros((n, n), order='F', dtype=complex)
    for j in range(n - 1):
        for i in range(j + 1, n):
            a[i, j] = np.random.randn() + 1j * np.random.randn()
            a[j, i] = -np.conj(a[i, j])
    for i in range(n):
        a[i, i] = 1j * np.random.randn()

    expected = np.max(np.abs(a))

    norm = ma02mz('M', 'L', a)

    np.testing.assert_allclose(norm, expected, rtol=1e-14)


def test_ma02mz_one_norm_upper():
    """
    Test one norm (1/O) with upper triangular storage.

    The one norm of a matrix is the maximum column sum of absolute values.
    For skew-Hermitian matrices, this equals the infinity norm.
    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    n = 4

    a = np.zeros((n, n), order='F', dtype=complex)
    for j in range(1, n):
        for i in range(j):
            a[i, j] = np.random.randn() + 1j * np.random.randn()
            a[j, i] = -np.conj(a[i, j])
    for i in range(n):
        a[i, i] = 1j * np.random.randn()

    expected = np.linalg.norm(a, 1)

    norm = ma02mz('1', 'U', a)

    np.testing.assert_allclose(norm, expected, rtol=1e-14)


def test_ma02mz_one_norm_lower():
    """
    Test one norm with lower triangular storage.

    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)
    n = 5

    a = np.zeros((n, n), order='F', dtype=complex)
    for j in range(n - 1):
        for i in range(j + 1, n):
            a[i, j] = np.random.randn() + 1j * np.random.randn()
            a[j, i] = -np.conj(a[i, j])
    for i in range(n):
        a[i, i] = 1j * np.random.randn()

    expected = np.linalg.norm(a, 1)

    norm = ma02mz('O', 'L', a)

    np.testing.assert_allclose(norm, expected, rtol=1e-14)


def test_ma02mz_infinity_norm():
    """
    Test infinity norm (I).

    For skew-Hermitian matrices, infinity norm equals one norm.
    Random seed: 111 (for reproducibility)
    """
    np.random.seed(111)
    n = 4

    a = np.zeros((n, n), order='F', dtype=complex)
    for j in range(1, n):
        for i in range(j):
            a[i, j] = np.random.randn() + 1j * np.random.randn()
            a[j, i] = -np.conj(a[i, j])
    for i in range(n):
        a[i, i] = 1j * np.random.randn()

    expected = np.linalg.norm(a, np.inf)

    norm = ma02mz('I', 'U', a)

    np.testing.assert_allclose(norm, expected, rtol=1e-14)


def test_ma02mz_frobenius_norm_upper():
    """
    Test Frobenius norm (F/E) with upper triangular storage.

    Frobenius norm is sqrt(sum of squares of all elements).
    Random seed: 222 (for reproducibility)
    """
    np.random.seed(222)
    n = 4

    a = np.zeros((n, n), order='F', dtype=complex)
    for j in range(1, n):
        for i in range(j):
            a[i, j] = np.random.randn() + 1j * np.random.randn()
            a[j, i] = -np.conj(a[i, j])
    for i in range(n):
        a[i, i] = 1j * np.random.randn()

    expected = np.linalg.norm(a, 'fro')

    norm = ma02mz('F', 'U', a)

    np.testing.assert_allclose(norm, expected, rtol=1e-13)


def test_ma02mz_frobenius_norm_lower():
    """
    Test Frobenius norm with lower triangular storage.

    Random seed: 333 (for reproducibility)
    """
    np.random.seed(333)
    n = 5

    a = np.zeros((n, n), order='F', dtype=complex)
    for j in range(n - 1):
        for i in range(j + 1, n):
            a[i, j] = np.random.randn() + 1j * np.random.randn()
            a[j, i] = -np.conj(a[i, j])
    for i in range(n):
        a[i, i] = 1j * np.random.randn()

    expected = np.linalg.norm(a, 'fro')

    norm = ma02mz('E', 'L', a)

    np.testing.assert_allclose(norm, expected, rtol=1e-13)


def test_ma02mz_n_zero():
    """
    Edge case: n = 0 should return 0.
    """
    a = np.array([], order='F', dtype=complex).reshape(0, 0)

    norm = ma02mz('F', 'U', a)

    assert norm == 0.0


def test_ma02mz_skew_hermitian_property():
    """
    Mathematical property: verify norm equality for skew-Hermitian matrices.

    For skew-Hermitian matrices:
    - ||A||_1 = ||A||_inf (one norm equals infinity norm)

    Random seed: 444 (for reproducibility)
    """
    np.random.seed(444)
    n = 6

    a = np.zeros((n, n), order='F', dtype=complex)
    for j in range(1, n):
        for i in range(j):
            a[i, j] = np.random.randn() + 1j * np.random.randn()
            a[j, i] = -np.conj(a[i, j])
    for i in range(n):
        a[i, i] = 1j * np.random.randn()

    norm_1 = ma02mz('1', 'U', a)
    norm_inf = ma02mz('I', 'U', a)

    np.testing.assert_allclose(norm_1, norm_inf, rtol=1e-14)


def test_ma02mz_simple_known_matrix():
    """
    Test with a simple known skew-Hermitian matrix.

    A = [  2i,       1+i,    -2+3i]
        [ -1+i,      -i,      4-i ]
        [  2+3i,   -4-i,      3i  ]

    Verify A is skew-Hermitian: A^H = -A
    - a[0,0] = 2i, -conj(a[0,0]) = -(-2i) = 2i, pure imaginary: OK
    - a[0,1] = 1+i, a[1,0] = -1+i = -(1-i) = -conj(1+i): OK
    - a[0,2] = -2+3i, a[2,0] = 2+3i = -(-2-3i) = -conj(-2+3i): OK
    - etc.

    Max norm: max(|2i|, |1+i|, |-2+3i|, |-i|, |4-i|, |3i|)
            = max(2, sqrt(2), sqrt(13), 1, sqrt(17), 3)
            = sqrt(17) approx 4.123

    One norm: max of column sums of absolute values
    """
    a = np.array([
        [2j,        1+1j,   -2+3j],
        [-1+1j,     -1j,     4-1j],
        [2+3j,      -4-1j,   3j]
    ], order='F', dtype=complex)

    np.testing.assert_allclose(a + a.conj().T, 0, atol=1e-14)

    norm_m = ma02mz('M', 'U', a)
    norm_1 = ma02mz('1', 'U', a)
    norm_f = ma02mz('F', 'U', a)

    expected_max = np.max(np.abs(a))
    expected_1 = np.linalg.norm(a, 1)
    expected_f = np.linalg.norm(a, 'fro')

    np.testing.assert_allclose(norm_m, expected_max, rtol=1e-14)
    np.testing.assert_allclose(norm_1, expected_1, rtol=1e-14)
    np.testing.assert_allclose(norm_f, expected_f, rtol=1e-13)


def test_ma02mz_diagonal_imaginary_only():
    """
    Test that diagonal contribution comes only from imaginary parts.

    For skew-Hermitian matrices, diagonal elements are pure imaginary.
    The routine assumes real parts of diagonal elements are zero.

    Random seed: 555 (for reproducibility)
    """
    np.random.seed(555)
    n = 3

    a = np.zeros((n, n), order='F', dtype=complex)
    for i in range(n):
        a[i, i] = 1j * np.random.randn()

    for j in range(1, n):
        for i in range(j):
            a[i, j] = np.random.randn() + 1j * np.random.randn()
            a[j, i] = -np.conj(a[i, j])

    norm_m = ma02mz('M', 'U', a)
    expected = np.max(np.abs(a))

    np.testing.assert_allclose(norm_m, expected, rtol=1e-14)



"""
Tests for MB03KA: Moving diagonal blocks in generalized periodic Schur form.

MB03KA reorders the diagonal blocks of a formal matrix product
T22_K^S(K) * T22_K-1^S(K-1) * ... * T22_1^S(1) of length K
by moving a block from row index IFST to row index ILST.
"""

import numpy as np
import pytest
from slicot import mb03ka


class TestMB03KABasic:
    """Basic functionality tests for MB03KA."""

    def test_move_block_forward_k2(self):
        """
        Test moving 1x1 diagonal block forward (IFST=1, ILST=2) with K=2.

        Random seed: 42 (for reproducibility)
        """
        np.random.seed(42)
        k = 2
        nc = 3
        kschur = 1
        ifst = 1
        ilst = 3

        n = np.array([nc, nc], dtype=np.int32)
        ni = np.array([0, 0], dtype=np.int32)
        s = np.array([1, 1], dtype=np.int32)

        t1 = np.array([
            [2.0, 0.5, 0.3],
            [0.0, 3.0, 0.4],
            [0.0, 0.0, 4.0]
        ], dtype=np.float64, order='F')

        t2 = np.array([
            [1.5, 0.2, 0.1],
            [0.0, 2.5, 0.3],
            [0.0, 0.0, 3.5]
        ], dtype=np.float64, order='F')

        t = np.concatenate([t1.ravel('F'), t2.ravel('F')])
        ldt = np.array([nc, nc], dtype=np.int32)
        ixt = np.array([1, nc*nc + 1], dtype=np.int32)

        q = np.eye(nc, dtype=np.float64, order='F').ravel('F')
        q = np.concatenate([q, np.eye(nc, dtype=np.float64, order='F').ravel('F')])
        ldq = np.array([nc, nc], dtype=np.int32)
        ixq = np.array([1, nc*nc + 1], dtype=np.int32)
        whichq = np.array([1, 1], dtype=np.int32)

        eps = np.finfo(np.float64).eps
        smlnum = np.finfo(np.float64).tiny / eps
        tol = np.array([10.0, eps, smlnum], dtype=np.float64)

        t_out, q_out, ifst_out, ilst_out, info = mb03ka(
            'U', whichq, False, k, nc, kschur, ifst, ilst,
            n, ni, s, t.copy(), ldt, ixt, q.copy(), ldq, ixq, tol
        )

        assert info == 0 or info == 1

    def test_move_block_backward_k2(self):
        """
        Test moving 1x1 diagonal block backward (IFST=3, ILST=1) with K=2.

        Random seed: 123 (for reproducibility)
        """
        np.random.seed(123)
        k = 2
        nc = 3
        kschur = 1
        ifst = 3
        ilst = 1

        n = np.array([nc, nc], dtype=np.int32)
        ni = np.array([0, 0], dtype=np.int32)
        s = np.array([1, 1], dtype=np.int32)

        t1 = np.array([
            [2.0, 0.5, 0.3],
            [0.0, 3.0, 0.4],
            [0.0, 0.0, 4.0]
        ], dtype=np.float64, order='F')

        t2 = np.array([
            [1.5, 0.2, 0.1],
            [0.0, 2.5, 0.3],
            [0.0, 0.0, 3.5]
        ], dtype=np.float64, order='F')

        t = np.concatenate([t1.ravel('F'), t2.ravel('F')])
        ldt = np.array([nc, nc], dtype=np.int32)
        ixt = np.array([1, nc*nc + 1], dtype=np.int32)

        q = np.eye(nc, dtype=np.float64, order='F').ravel('F')
        q = np.concatenate([q, np.eye(nc, dtype=np.float64, order='F').ravel('F')])
        ldq = np.array([nc, nc], dtype=np.int32)
        ixq = np.array([1, nc*nc + 1], dtype=np.int32)
        whichq = np.array([1, 1], dtype=np.int32)

        eps = np.finfo(np.float64).eps
        smlnum = np.finfo(np.float64).tiny / eps
        tol = np.array([10.0, eps, smlnum], dtype=np.float64)

        t_out, q_out, ifst_out, ilst_out, info = mb03ka(
            'U', whichq, False, k, nc, kschur, ifst, ilst,
            n, ni, s, t.copy(), ldt, ixt, q.copy(), ldq, ixq, tol
        )

        assert info == 0 or info == 1

    def test_move_2x2_block(self):
        """
        Test moving a 2x2 diagonal block with K=2.

        Tests IFST pointing to row 1 of a 2x2 block, moving to position 3.
        Random seed: 456 (for reproducibility)
        """
        np.random.seed(456)
        k = 2
        nc = 4
        kschur = 1
        ifst = 1
        ilst = 3

        n = np.array([nc, nc], dtype=np.int32)
        ni = np.array([0, 0], dtype=np.int32)
        s = np.array([1, 1], dtype=np.int32)

        t1 = np.array([
            [1.0, 0.5, 0.2, 0.1],
            [0.1, 1.2, 0.3, 0.2],
            [0.0, 0.0, 2.0, 0.4],
            [0.0, 0.0, 0.0, 3.0]
        ], dtype=np.float64, order='F')

        t2 = np.array([
            [0.8, 0.3, 0.1, 0.05],
            [0.05, 0.9, 0.2, 0.1],
            [0.0, 0.0, 1.5, 0.3],
            [0.0, 0.0, 0.0, 2.5]
        ], dtype=np.float64, order='F')

        t = np.concatenate([t1.ravel('F'), t2.ravel('F')])
        ldt = np.array([nc, nc], dtype=np.int32)
        ixt = np.array([1, nc*nc + 1], dtype=np.int32)

        q = np.eye(nc, dtype=np.float64, order='F').ravel('F')
        q = np.concatenate([q, np.eye(nc, dtype=np.float64, order='F').ravel('F')])
        ldq = np.array([nc, nc], dtype=np.int32)
        ixq = np.array([1, nc*nc + 1], dtype=np.int32)
        whichq = np.array([1, 1], dtype=np.int32)

        eps = np.finfo(np.float64).eps
        smlnum = np.finfo(np.float64).tiny / eps
        tol = np.array([10.0, eps, smlnum], dtype=np.float64)

        t_out, q_out, ifst_out, ilst_out, info = mb03ka(
            'U', whichq, True, k, nc, kschur, ifst, ilst,
            n, ni, s, t.copy(), ldt, ixt, q.copy(), ldq, ixq, tol
        )

        assert info == 0 or info == 1


class TestMB03KANoQ:
    """Tests with COMPQ='N' (no Q update)."""

    def test_no_q_computation(self):
        """
        Test with COMPQ='N' - no orthogonal matrices computed.

        Random seed: 111 (for reproducibility)
        """
        np.random.seed(111)
        k = 2
        nc = 3
        kschur = 1
        ifst = 1
        ilst = 2

        n = np.array([nc, nc], dtype=np.int32)
        ni = np.array([0, 0], dtype=np.int32)
        s = np.array([1, 1], dtype=np.int32)

        t1 = np.array([
            [2.0, 0.5, 0.3],
            [0.0, 3.0, 0.4],
            [0.0, 0.0, 4.0]
        ], dtype=np.float64, order='F')

        t2 = np.array([
            [1.5, 0.2, 0.1],
            [0.0, 2.5, 0.3],
            [0.0, 0.0, 3.5]
        ], dtype=np.float64, order='F')

        t = np.concatenate([t1.ravel('F'), t2.ravel('F')])
        ldt = np.array([nc, nc], dtype=np.int32)
        ixt = np.array([1, nc*nc + 1], dtype=np.int32)

        q = np.zeros(1, dtype=np.float64)
        ldq = np.array([1, 1], dtype=np.int32)
        ixq = np.array([1, 1], dtype=np.int32)
        whichq = np.array([0, 0], dtype=np.int32)

        eps = np.finfo(np.float64).eps
        smlnum = np.finfo(np.float64).tiny / eps
        tol = np.array([10.0, eps, smlnum], dtype=np.float64)

        t_out, q_out, ifst_out, ilst_out, info = mb03ka(
            'N', whichq, False, k, nc, kschur, ifst, ilst,
            n, ni, s, t.copy(), ldt, ixt, q.copy(), ldq, ixq, tol
        )

        assert info == 0 or info == 1


class TestMB03KAWithSignature:
    """Tests with different signature arrays S."""

    def test_mixed_signatures(self):
        """
        Test with mixed signatures S=[1, -1].

        Random seed: 222 (for reproducibility)
        """
        np.random.seed(222)
        k = 2
        nc = 3
        kschur = 1
        ifst = 1
        ilst = 2

        n = np.array([nc, nc], dtype=np.int32)
        ni = np.array([0, 0], dtype=np.int32)
        s = np.array([1, -1], dtype=np.int32)

        t1 = np.array([
            [2.0, 0.5, 0.3],
            [0.0, 3.0, 0.4],
            [0.0, 0.0, 4.0]
        ], dtype=np.float64, order='F')

        t2 = np.array([
            [1.5, 0.2, 0.1],
            [0.0, 2.5, 0.3],
            [0.0, 0.0, 3.5]
        ], dtype=np.float64, order='F')

        t = np.concatenate([t1.ravel('F'), t2.ravel('F')])
        ldt = np.array([nc, nc], dtype=np.int32)
        ixt = np.array([1, nc*nc + 1], dtype=np.int32)

        q = np.eye(nc, dtype=np.float64, order='F').ravel('F')
        q = np.concatenate([q, np.eye(nc, dtype=np.float64, order='F').ravel('F')])
        ldq = np.array([nc, nc], dtype=np.int32)
        ixq = np.array([1, nc*nc + 1], dtype=np.int32)
        whichq = np.array([1, 1], dtype=np.int32)

        eps = np.finfo(np.float64).eps
        smlnum = np.finfo(np.float64).tiny / eps
        tol = np.array([10.0, eps, smlnum], dtype=np.float64)

        t_out, q_out, ifst_out, ilst_out, info = mb03ka(
            'U', whichq, False, k, nc, kschur, ifst, ilst,
            n, ni, s, t.copy(), ldt, ixt, q.copy(), ldq, ixq, tol
        )

        assert info == 0 or info == 1


class TestMB03KAWorkspaceQuery:
    """Test workspace query functionality."""

    def test_workspace_query(self):
        """
        Test workspace query with LDWORK=-1.

        Random seed: 333 (for reproducibility)
        """
        np.random.seed(333)
        k = 2
        nc = 3
        kschur = 1
        ifst = 1
        ilst = 2

        n = np.array([nc, nc], dtype=np.int32)
        ni = np.array([0, 0], dtype=np.int32)
        s = np.array([1, 1], dtype=np.int32)

        t1 = np.array([
            [2.0, 0.5, 0.3],
            [0.0, 3.0, 0.4],
            [0.0, 0.0, 4.0]
        ], dtype=np.float64, order='F')

        t2 = np.array([
            [1.5, 0.2, 0.1],
            [0.0, 2.5, 0.3],
            [0.0, 0.0, 3.5]
        ], dtype=np.float64, order='F')

        t = np.concatenate([t1.ravel('F'), t2.ravel('F')])
        ldt = np.array([nc, nc], dtype=np.int32)
        ixt = np.array([1, nc*nc + 1], dtype=np.int32)

        q = np.zeros(1, dtype=np.float64)
        ldq = np.array([1, 1], dtype=np.int32)
        ixq = np.array([1, 1], dtype=np.int32)
        whichq = np.array([0, 0], dtype=np.int32)

        eps = np.finfo(np.float64).eps
        smlnum = np.finfo(np.float64).tiny / eps
        tol = np.array([10.0, eps, smlnum], dtype=np.float64)

        t_out, q_out, ifst_out, ilst_out, info = mb03ka(
            'N', whichq, False, k, nc, kschur, ifst, ilst,
            n, ni, s, t.copy(), ldt, ixt, q.copy(), ldq, ixq, tol,
            ldwork=-1
        )

        assert info == 0


class TestMB03KAQuickReturn:
    """Test quick return conditions."""

    def test_ifst_eq_ilst(self):
        """Test quick return when IFST equals ILST (no reordering needed)."""
        k = 2
        nc = 3
        kschur = 1
        ifst = 2
        ilst = 2

        n = np.array([nc, nc], dtype=np.int32)
        ni = np.array([0, 0], dtype=np.int32)
        s = np.array([1, 1], dtype=np.int32)

        t = np.eye(nc, dtype=np.float64, order='F').ravel('F')
        t = np.concatenate([t, np.eye(nc, dtype=np.float64, order='F').ravel('F')])
        ldt = np.array([nc, nc], dtype=np.int32)
        ixt = np.array([1, nc*nc + 1], dtype=np.int32)

        q = np.zeros(1, dtype=np.float64)
        ldq = np.array([1, 1], dtype=np.int32)
        ixq = np.array([1, 1], dtype=np.int32)
        whichq = np.array([0, 0], dtype=np.int32)

        eps = np.finfo(np.float64).eps
        smlnum = np.finfo(np.float64).tiny / eps
        tol = np.array([10.0, eps, smlnum], dtype=np.float64)

        t_out, q_out, ifst_out, ilst_out, info = mb03ka(
            'N', whichq, False, k, nc, kschur, ifst, ilst,
            n, ni, s, t.copy(), ldt, ixt, q.copy(), ldq, ixq, tol
        )

        assert info == 0


class TestMB03KAOrthogonality:
    """Test orthogonality of computed Q matrices."""

    def test_q_orthogonality(self):
        """
        Verify Q matrices are orthogonal after reordering.

        Random seed: 444 (for reproducibility)
        """
        np.random.seed(444)
        k = 2
        nc = 3
        kschur = 1
        ifst = 1
        ilst = 3

        n = np.array([nc, nc], dtype=np.int32)
        ni = np.array([0, 0], dtype=np.int32)
        s = np.array([1, 1], dtype=np.int32)

        t1 = np.array([
            [2.0, 0.5, 0.3],
            [0.0, 3.0, 0.4],
            [0.0, 0.0, 4.0]
        ], dtype=np.float64, order='F')

        t2 = np.array([
            [1.5, 0.2, 0.1],
            [0.0, 2.5, 0.3],
            [0.0, 0.0, 3.5]
        ], dtype=np.float64, order='F')

        t = np.concatenate([t1.ravel('F'), t2.ravel('F')])
        ldt = np.array([nc, nc], dtype=np.int32)
        ixt = np.array([1, nc*nc + 1], dtype=np.int32)

        q = np.eye(nc, dtype=np.float64, order='F').ravel('F')
        q = np.concatenate([q, np.eye(nc, dtype=np.float64, order='F').ravel('F')])
        ldq = np.array([nc, nc], dtype=np.int32)
        ixq = np.array([1, nc*nc + 1], dtype=np.int32)
        whichq = np.array([1, 1], dtype=np.int32)

        eps = np.finfo(np.float64).eps
        smlnum = np.finfo(np.float64).tiny / eps
        tol = np.array([10.0, eps, smlnum], dtype=np.float64)

        t_out, q_out, ifst_out, ilst_out, info = mb03ka(
            'U', whichq, False, k, nc, kschur, ifst, ilst,
            n, ni, s, t.copy(), ldt, ixt, q.copy(), ldq, ixq, tol
        )

        if info == 0:
            q1 = q_out[:nc*nc].reshape((nc, nc), order='F')
            q2 = q_out[nc*nc:].reshape((nc, nc), order='F')

            np.testing.assert_allclose(q1.T @ q1, np.eye(nc), rtol=1e-12, atol=1e-14)
            np.testing.assert_allclose(q2.T @ q2, np.eye(nc), rtol=1e-12, atol=1e-14)


class TestMB03KASelectiveQ:
    """Tests with COMPQ='W' (selective Q computation)."""

    def test_selective_q(self):
        """
        Test with COMPQ='W' - selective Q computation via WHICHQ.

        Random seed: 555 (for reproducibility)
        """
        np.random.seed(555)
        k = 2
        nc = 3
        kschur = 1
        ifst = 1
        ilst = 2

        n = np.array([nc, nc], dtype=np.int32)
        ni = np.array([0, 0], dtype=np.int32)
        s = np.array([1, 1], dtype=np.int32)

        t1 = np.array([
            [2.0, 0.5, 0.3],
            [0.0, 3.0, 0.4],
            [0.0, 0.0, 4.0]
        ], dtype=np.float64, order='F')

        t2 = np.array([
            [1.5, 0.2, 0.1],
            [0.0, 2.5, 0.3],
            [0.0, 0.0, 3.5]
        ], dtype=np.float64, order='F')

        t = np.concatenate([t1.ravel('F'), t2.ravel('F')])
        ldt = np.array([nc, nc], dtype=np.int32)
        ixt = np.array([1, nc*nc + 1], dtype=np.int32)

        q = np.eye(nc, dtype=np.float64, order='F').ravel('F')
        q = np.concatenate([q, np.eye(nc, dtype=np.float64, order='F').ravel('F')])
        ldq = np.array([nc, nc], dtype=np.int32)
        ixq = np.array([1, nc*nc + 1], dtype=np.int32)
        whichq = np.array([1, 0], dtype=np.int32)

        eps = np.finfo(np.float64).eps
        smlnum = np.finfo(np.float64).tiny / eps
        tol = np.array([10.0, eps, smlnum], dtype=np.float64)

        t_out, q_out, ifst_out, ilst_out, info = mb03ka(
            'W', whichq, False, k, nc, kschur, ifst, ilst,
            n, ni, s, t.copy(), ldt, ixt, q.copy(), ldq, ixq, tol
        )

        assert info == 0 or info == 1


if __name__ == '__main__':
    pytest.main([__file__, '-v'])



"""
Tests for MB03YA - Annihilate subdiagonal entries of Hessenberg matrix A.

MB03YA annihilates one or two entries on the subdiagonal of the Hessenberg
matrix A for dealing with zero elements on the diagonal of the triangular
matrix B. This is an auxiliary routine for MB03XP and MB03YD.

Test data sources:
- Mathematical properties of QR/RQ steps
- Validation against known transformations
"""

import numpy as np
import pytest

from slicot import mb03ya


def test_mb03ya_basic():
    """
    Test basic functionality with small matrix where B has zero diagonal.

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n = 4
    ilo, ihi = 1, 4
    iloq, ihiq = 1, 4
    pos = 2

    a = np.zeros((n, n), order='F', dtype=float)
    for i in range(n):
        for j in range(i, n):
            a[i, j] = np.random.randn()
        if i < n - 1:
            a[i + 1, i] = np.random.randn()

    b = np.zeros((n, n), order='F', dtype=float)
    for i in range(n):
        for j in range(i, n):
            if i != pos - 1 or j != pos - 1:
                b[i, j] = np.random.randn()
    b[pos - 1, pos - 1] = 0.0

    q = np.eye(n, order='F', dtype=float)
    z = np.eye(n, order='F', dtype=float)

    a_out, b_out, q_out, z_out, info = mb03ya(
        True, True, True, ilo, ihi, iloq, ihiq, pos, a, b, q, z
    )

    assert info == 0
    if pos > ilo:
        assert abs(a_out[pos - 1, pos - 2]) < 1e-14
    if pos < ihi:
        assert abs(a_out[pos, pos - 1]) < 1e-14


def test_mb03ya_no_accumulate():
    """
    Test without accumulating Q and Z.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n = 3
    ilo, ihi = 1, 3
    iloq, ihiq = 1, 3
    pos = 2

    a = np.array([
        [1.0, 2.0, 3.0],
        [4.0, 5.0, 6.0],
        [0.0, 7.0, 8.0]
    ], order='F', dtype=float)

    b = np.array([
        [1.0, 2.0, 3.0],
        [0.0, 0.0, 4.0],
        [0.0, 0.0, 5.0]
    ], order='F', dtype=float)

    q = np.eye(n, order='F', dtype=float)
    z = np.eye(n, order='F', dtype=float)

    a_out, b_out, q_out, z_out, info = mb03ya(
        True, False, False, ilo, ihi, iloq, ihiq, pos, a, b, q, z
    )

    assert info == 0


def test_mb03ya_boundary_pos():
    """
    Test with pos at the boundary (first position).

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    n = 3
    ilo, ihi = 1, 3
    iloq, ihiq = 1, 3
    pos = 1

    a = np.array([
        [1.0, 2.0, 3.0],
        [4.0, 5.0, 6.0],
        [0.0, 7.0, 8.0]
    ], order='F', dtype=float)

    b = np.array([
        [0.0, 2.0, 3.0],
        [0.0, 4.0, 5.0],
        [0.0, 0.0, 6.0]
    ], order='F', dtype=float)

    q = np.eye(n, order='F', dtype=float)
    z = np.eye(n, order='F', dtype=float)

    a_out, b_out, q_out, z_out, info = mb03ya(
        True, True, True, ilo, ihi, iloq, ihiq, pos, a, b, q, z
    )

    assert info == 0


def test_mb03ya_invalid_pos():
    """Test error handling for invalid pos parameter."""
    n = 3
    ilo, ihi = 1, 3
    iloq, ihiq = 1, 3
    pos = 0

    a = np.eye(n, order='F', dtype=float)
    b = np.eye(n, order='F', dtype=float)
    q = np.eye(n, order='F', dtype=float)
    z = np.eye(n, order='F', dtype=float)

    a_out, b_out, q_out, z_out, info = mb03ya(
        True, True, True, ilo, ihi, iloq, ihiq, pos, a, b, q, z
    )

    assert info == -9


def test_mb03ya_orthogonality():
    """
    Test that Q and Z remain orthogonal after transformation.

    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)
    n = 4
    ilo, ihi = 1, 4
    iloq, ihiq = 1, 4
    pos = 2

    a = np.zeros((n, n), order='F', dtype=float)
    for i in range(n):
        for j in range(i, n):
            a[i, j] = np.random.randn()
        if i < n - 1:
            a[i + 1, i] = np.random.randn()

    b = np.zeros((n, n), order='F', dtype=float)
    for i in range(n):
        for j in range(i, n):
            if i != pos - 1 or j != pos - 1:
                b[i, j] = np.random.randn()
    b[pos - 1, pos - 1] = 0.0

    q = np.eye(n, order='F', dtype=float)
    z = np.eye(n, order='F', dtype=float)

    a_out, b_out, q_out, z_out, info = mb03ya(
        True, True, True, ilo, ihi, iloq, ihiq, pos, a, b, q, z
    )

    assert info == 0
    np.testing.assert_allclose(q_out @ q_out.T, np.eye(n), rtol=1e-14, atol=1e-14)
    np.testing.assert_allclose(z_out @ z_out.T, np.eye(n), rtol=1e-14, atol=1e-14)


def test_mb03ya_upper_triangular_b():
    """
    Test that B remains upper triangular after transformation.

    Random seed: 999 (for reproducibility)
    """
    np.random.seed(999)
    n = 4
    ilo, ihi = 1, 4
    iloq, ihiq = 1, 4
    pos = 3

    a = np.zeros((n, n), order='F', dtype=float)
    for i in range(n):
        for j in range(i, n):
            a[i, j] = np.random.randn()
        if i < n - 1:
            a[i + 1, i] = np.random.randn()

    b = np.zeros((n, n), order='F', dtype=float)
    for i in range(n):
        for j in range(i, n):
            if i != pos - 1 or j != pos - 1:
                b[i, j] = np.random.randn()
    b[pos - 1, pos - 1] = 0.0

    q = np.eye(n, order='F', dtype=float)
    z = np.eye(n, order='F', dtype=float)

    a_out, b_out, q_out, z_out, info = mb03ya(
        True, True, True, ilo, ihi, iloq, ihiq, pos, a, b, q, z
    )

    assert info == 0
    for i in range(1, n):
        for j in range(i):
            assert abs(b_out[i, j]) < 1e-14, f"B[{i},{j}] = {b_out[i,j]} not zero"



"""
Tests for AB01ND - Controllable realization for multi-input systems.

AB01ND finds a controllable realization for the linear time-invariant
multi-input system dX/dt = A * X + B * U, reducing (A,B) to orthogonal
canonical form where Acont is upper block Hessenberg with full row rank
subdiagonal blocks.
"""

import numpy as np
import pytest
from slicot import ab01nd


"""Basic functionality tests from HTML documentation example."""

def test_html_doc_example():
    """
    Test case from SLICOT HTML documentation.

    Input: N=3, M=2, TOL=0.0, JOBZ='I'
    A read row-wise:
        -1.0   0.0   0.0
        -2.0  -2.0  -2.0
        -1.0   0.0  -3.0
    B read column-wise:
         1.0   0.0   0.0
         0.0   2.0   1.0

    Expected NCONT=2
    Expected INDCON=1
    Expected NBLK=[2]
    Expected Acont:
        -3.0000   2.2361
         0.0000  -1.0000
    Expected Bcont:
         0.0000  -2.2361
         1.0000   0.0000
    Expected Z:
         0.0000   1.0000   0.0000
        -0.8944   0.0000  -0.4472
        -0.4472   0.0000   0.8944
    """
    n, m = 3, 2

    # A read row-wise from HTML (standard row order)
    a = np.array([
        [-1.0,  0.0,  0.0],
        [-2.0, -2.0, -2.0],
        [-1.0,  0.0, -3.0]
    ], dtype=float, order='F')

    # B read column-wise: ((B(I,J), I=1,N), J=1,M)
    # First column: 1.0, 0.0, 0.0
    # Second column: 0.0, 2.0, 1.0
    b = np.array([
        [1.0, 0.0],
        [0.0, 2.0],
        [0.0, 1.0]
    ], dtype=float, order='F')

    tol = 0.0

    a_out, b_out, ncont, indcon, nblk, z, tau, info = ab01nd('I', a, b, tol)

    assert info == 0
    assert ncont == 2
    assert indcon == 1

    # Expected nblk has indcon=1 element
    assert nblk[0] == 2

    # Expected Acont (NCONT x NCONT)
    a_expected = np.array([
        [-3.0000,  2.2361],
        [ 0.0000, -1.0000]
    ], dtype=float, order='F')

    # Expected Bcont (NCONT x M)
    b_expected = np.array([
        [0.0000, -2.2361],
        [1.0000,  0.0000]
    ], dtype=float, order='F')

    # Expected Z (N x N)
    z_expected = np.array([
        [ 0.0000,  1.0000,  0.0000],
        [-0.8944,  0.0000, -0.4472],
        [-0.4472,  0.0000,  0.8944]
    ], dtype=float, order='F')

    # Validate outputs - HTML doc shows 4 decimal places
    np.testing.assert_allclose(a_out[:ncont, :ncont], a_expected, rtol=1e-3, atol=1e-4)
    np.testing.assert_allclose(b_out[:ncont, :], b_expected, rtol=1e-3, atol=1e-4)
    np.testing.assert_allclose(z, z_expected, rtol=1e-3, atol=1e-4)


"""Test mathematical properties of orthogonal transformation."""

def test_z_is_orthogonal():
    """
    Validate Z'*Z = I (orthogonality property).

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n, m = 4, 2

    a = np.random.randn(n, n).astype(float, order='F')
    b = np.random.randn(n, m).astype(float, order='F')
    b[0, 0] = 1.0  # Ensure non-zero for controllability

    _, _, ncont, indcon, nblk, z, _, info = ab01nd('I', a, b, 0.0)

    assert info == 0

    # Z should be orthogonal: Z'*Z = I
    ztz = z.T @ z
    np.testing.assert_allclose(ztz, np.eye(n), rtol=1e-14, atol=1e-14)

def test_similarity_transformation():
    """
    Validate A_out = Z' * A_in * Z (similarity transformation property).

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n, m = 5, 2

    a_in = np.random.randn(n, n).astype(float, order='F')
    b_in = np.random.randn(n, m).astype(float, order='F')
    b_in[0, 0] = 2.0

    a_copy = a_in.copy()

    a_out, _, ncont, _, _, z, _, info = ab01nd('I', a_in, b_in, 0.0)

    assert info == 0

    # A_out = Z' * A_orig * Z
    a_transformed = z.T @ a_copy @ z
    np.testing.assert_allclose(a_out, a_transformed, rtol=1e-13, atol=1e-14)

def test_b_transformation():
    """
    Validate B_out = Z' * B_in (matrix transformation).

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    n, m = 4, 3

    a = np.random.randn(n, n).astype(float, order='F')
    b_in = np.random.randn(n, m).astype(float, order='F')
    b_in[0, 0] = 1.5

    b_copy = b_in.copy()

    _, b_out, ncont, _, _, z, _, info = ab01nd('I', a, b_in, 0.0)

    assert info == 0

    # B_out = Z' * B_orig
    b_transformed = z.T @ b_copy
    np.testing.assert_allclose(b_out, b_transformed, rtol=1e-13, atol=1e-14)


"""Test upper block Hessenberg structure of output A."""

def test_output_is_block_hessenberg():
    """
    Validate output Acont has upper block Hessenberg structure.

    For block Hessenberg with blocks defined by NBLK, elements below
    the first block subdiagonal should be zero.

    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)
    n, m = 6, 3

    a = np.random.randn(n, n).astype(float, order='F')
    b = np.random.randn(n, m).astype(float, order='F')
    b[0, 0] = 3.0

    a_out, _, ncont, indcon, nblk, _, _, info = ab01nd('I', a, b, 0.0)

    assert info == 0

    # Build block boundaries from nblk
    block_ends = []
    pos = 0
    for i in range(indcon):
        pos += nblk[i]
        block_ends.append(pos)

    # Check zeros below block subdiagonal
    # For each column block j, rows from block j+2 and beyond should be zero
    row_start = 0
    for block_idx in range(indcon):
        block_size = nblk[block_idx]
        col_end = block_ends[block_idx]

        # Rows from block (block_idx + 2) to ncont should be zero in these columns
        if block_idx + 2 <= indcon - 1:
            row_check_start = block_ends[block_idx + 1] if block_idx + 1 < len(block_ends) else ncont
            for j in range(row_start, col_end):
                for i in range(row_check_start, ncont):
                    assert abs(a_out[i, j]) < 1e-12, \
                        f"Element ({i},{j}) should be zero: {a_out[i, j]}"
        row_start = col_end

def test_subdiagonal_blocks_full_rank():
    """
    Validate subdiagonal blocks have full row rank.

    Random seed: 888 (for reproducibility)
    """
    np.random.seed(888)
    n, m = 8, 4

    a = np.random.randn(n, n).astype(float, order='F')
    b = np.random.randn(n, m).astype(float, order='F')
    b[:, 0] = np.array([1.0, 0.5, 0.3, 0.1, 0.05, 0.02, 0.01, 0.005])

    a_out, _, ncont, indcon, nblk, _, _, info = ab01nd('I', a, b, 0.0)

    assert info == 0

    if indcon > 1:
        # Check each subdiagonal block A_{i+1,i}
        row_start = 0
        col_start = 0
        for i in range(indcon - 1):
            block_rows = nblk[i + 1]
            block_cols = nblk[i]
            row_start_next = row_start + nblk[i]

            subblock = a_out[row_start_next:row_start_next + block_rows,
                             col_start:col_start + block_cols]

            # Full row rank means rank == number of rows
            rank = np.linalg.matrix_rank(subblock)
            assert rank == block_rows, \
                f"Subdiagonal block {i} has rank {rank}, expected {block_rows}"

            col_start += block_cols
            row_start = row_start_next


"""Test controllability detection."""

def test_partially_controllable_system():
    """
    Test system that is not fully controllable.

    Create a block diagonal A where B only affects some blocks.
    """
    n, m = 4, 1

    # Block diagonal A with two 2x2 blocks
    a = np.array([
        [1.0, 2.0, 0.0, 0.0],
        [0.0, 1.0, 0.0, 0.0],
        [0.0, 0.0, 3.0, 4.0],
        [0.0, 0.0, 0.0, 3.0]
    ], dtype=float, order='F')

    # B only affects first block
    b = np.array([[1.0], [0.0], [0.0], [0.0]], dtype=float, order='F')

    _, _, ncont, indcon, nblk, _, _, info = ab01nd('I', a, b, 0.0)

    assert info == 0
    assert ncont < n  # Not fully controllable

def test_zero_b_matrix():
    """
    Test with zero B matrix - completely uncontrollable.
    """
    n, m = 3, 2

    a = np.array([
        [1.0, 2.0, 3.0],
        [4.0, 5.0, 6.0],
        [7.0, 8.0, 9.0]
    ], dtype=float, order='F')

    b = np.zeros((n, m), dtype=float, order='F')

    _, _, ncont, indcon, nblk, z, _, info = ab01nd('I', a, b, 0.0)

    assert info == 0
    assert ncont == 0
    assert indcon == 0
    # For zero B with JOBZ='I', Z should be identity
    np.testing.assert_allclose(z, np.eye(n), rtol=1e-14, atol=1e-14)

def test_fully_controllable_random_system():
    """
    Test that random system is typically fully controllable.

    Random seed: 333 (for reproducibility)
    """
    np.random.seed(333)
    n, m = 5, 3

    a = np.random.randn(n, n).astype(float, order='F')
    b = np.random.randn(n, m).astype(float, order='F')

    _, _, ncont, indcon, nblk, _, _, info = ab01nd('I', a, b, 0.0)

    assert info == 0
    assert ncont == n  # Fully controllable


"""Test different JOBZ modes."""

def test_jobz_n():
    """
    Test JOBZ='N' mode (no orthogonal transformation stored).
    """
    n, m = 3, 2

    a = np.array([
        [-1.0,  0.0,  0.0],
        [-2.0, -2.0, -2.0],
        [-1.0,  0.0, -3.0]
    ], dtype=float, order='F')

    b = np.array([
        [1.0, 0.0],
        [0.0, 2.0],
        [0.0, 1.0]
    ], dtype=float, order='F')

    a_out, b_out, ncont, indcon, nblk, z, tau, info = ab01nd('N', a, b, 0.0)

    assert info == 0
    assert ncont == 2
    assert indcon == 1

def test_jobz_f():
    """
    Test JOBZ='F' mode (factored form storage).
    """
    n, m = 3, 2

    a = np.array([
        [-1.0,  0.0,  0.0],
        [-2.0, -2.0, -2.0],
        [-1.0,  0.0, -3.0]
    ], dtype=float, order='F')

    b = np.array([
        [1.0, 0.0],
        [0.0, 2.0],
        [0.0, 1.0]
    ], dtype=float, order='F')

    a_out, b_out, ncont, indcon, nblk, z, tau, info = ab01nd('F', a, b, 0.0)

    assert info == 0
    assert ncont == 2
    assert indcon == 1


"""Test edge cases and boundary conditions."""

def test_n_equals_1_m_equals_1():
    """
    Test with N=1, M=1 (scalar system, single input).
    """
    a = np.array([[2.0]], dtype=float, order='F')
    b = np.array([[3.0]], dtype=float, order='F')

    a_out, b_out, ncont, indcon, nblk, z, tau, info = ab01nd('I', a, b, 0.0)

    assert info == 0
    assert ncont == 1
    assert indcon == 1
    assert nblk[0] == 1

def test_n_equals_1_m_equals_2():
    """
    Test with N=1, M=2 (scalar system, two inputs).
    """
    a = np.array([[2.0]], dtype=float, order='F')
    b = np.array([[3.0, 4.0]], dtype=float, order='F')

    a_out, b_out, ncont, indcon, nblk, z, tau, info = ab01nd('I', a, b, 0.0)

    assert info == 0
    assert ncont == 1
    assert indcon == 1

def test_n_equals_0():
    """
    Test with N=0 (empty system).
    """
    a = np.array([], dtype=float, order='F').reshape(0, 0)
    b = np.array([], dtype=float, order='F').reshape(0, 2)

    a_out, b_out, ncont, indcon, nblk, z, tau, info = ab01nd('I', a, b, 0.0)

    assert info == 0
    assert ncont == 0
    assert indcon == 0

def test_m_equals_0():
    """
    Test with M=0 (no inputs).
    """
    n = 3
    a = np.eye(n, dtype=float, order='F')
    b = np.array([], dtype=float, order='F').reshape(n, 0)

    a_out, b_out, ncont, indcon, nblk, z, tau, info = ab01nd('I', a, b, 0.0)

    assert info == 0
    assert ncont == 0
    assert indcon == 0
    # Z should be identity for JOBZ='I'
    np.testing.assert_allclose(z, np.eye(n), rtol=1e-14)

def test_large_system():
    """
    Test with larger system to verify scalability.

    Random seed: 999 (for reproducibility)
    """
    np.random.seed(999)
    n, m = 20, 5

    a = np.random.randn(n, n).astype(float, order='F')
    b = np.random.randn(n, m).astype(float, order='F')

    a_out, _, ncont, indcon, nblk, z, _, info = ab01nd('I', a, b, 0.0)

    assert info == 0

    # Z should be orthogonal
    ztz = z.T @ z
    np.testing.assert_allclose(ztz, np.eye(n), rtol=1e-13, atol=1e-13)


"""Test tolerance handling."""

def test_explicit_tolerance():
    """
    Test with explicit positive tolerance.
    """
    n, m = 3, 2

    a = np.array([
        [-1.0,  0.0,  0.0],
        [-2.0, -2.0, -2.0],
        [-1.0,  0.0, -3.0]
    ], dtype=float, order='F')

    b = np.array([
        [1.0, 0.0],
        [0.0, 2.0],
        [0.0, 1.0]
    ], dtype=float, order='F')

    tol = 1e-10

    a_out, b_out, ncont, indcon, nblk, z, tau, info = ab01nd('I', a, b, tol)

    assert info == 0
    assert ncont == 2
    assert indcon == 1

def test_negative_tolerance_uses_default():
    """
    Test that negative/zero tolerance uses default (N*N*EPS).
    """
    n, m = 3, 2

    a = np.array([
        [-1.0,  0.0,  0.0],
        [-2.0, -2.0, -2.0],
        [-1.0,  0.0, -3.0]
    ], dtype=float, order='F')

    b = np.array([
        [1.0, 0.0],
        [0.0, 2.0],
        [0.0, 1.0]
    ], dtype=float, order='F')

    # Zero tol means use default
    tol = 0.0

    a_out, b_out, ncont, indcon, nblk, z, tau, info = ab01nd('I', a, b, tol)

    assert info == 0
    assert ncont == 2


"""Test error handling."""

def test_invalid_jobz():
    """
    Test invalid JOBZ parameter.
    """
    n, m = 3, 2
    a = np.eye(n, dtype=float, order='F')
    b = np.ones((n, m), dtype=float, order='F')

    with pytest.raises(ValueError, match="[Jj]obz"):
        ab01nd('X', a, b, 0.0)

def test_mismatched_dimensions():
    """
    Test mismatched A and B dimensions.
    """
    a = np.eye(3, dtype=float, order='F')
    b = np.ones((4, 2), dtype=float, order='F')  # Wrong number of rows

    with pytest.raises(ValueError):
        ab01nd('I', a, b, 0.0)


"""Test eigenvalue preservation under similarity transformation."""

def test_eigenvalues_preserved():
    """
    Validate eigenvalues are preserved by similarity transformation.

    Property: eigenvalues of A equal eigenvalues of Z'*A*Z.

    Random seed: 555 (for reproducibility)
    """
    np.random.seed(555)
    n, m = 5, 2

    a = np.random.randn(n, n).astype(float, order='F')
    b = np.random.randn(n, m).astype(float, order='F')

    eig_before = np.linalg.eigvals(a)

    a_out, _, ncont, _, _, z, _, info = ab01nd('I', a, b, 0.0)

    assert info == 0

    eig_after = np.linalg.eigvals(a_out)

    # Sort by real then imaginary for comparison
    eig_before_sorted = np.sort_complex(eig_before)
    eig_after_sorted = np.sort_complex(eig_after)

    np.testing.assert_allclose(eig_before_sorted, eig_after_sorted,
                               rtol=1e-13, atol=1e-14)



"""
Tests for MB04RT - Blocked generalized real Sylvester equation solver.

Solves using Level 3 BLAS:
    A * R - L * B = scale * C
    D * R - L * E = scale * F

where (A, D) and (B, E) are in generalized Schur form (A, B upper quasi-triangular,
D, E upper triangular).

MB04RT is the blocked version of MB04RS, using optimal block sizes from ILAENV.
Solution is aborted early if any element exceeds PMAX.
"""

import numpy as np
import pytest
from slicot import mb04rt


def test_mb04rt_basic_1x1():
    """
    Test 1x1 system - smallest case, goes through unblocked path.

    Random seed: 42 (for reproducibility)
    """
    m, n = 1, 1
    pmax = 1e10

    a = np.array([[2.0]], order='F', dtype=float)
    b = np.array([[3.0]], order='F', dtype=float)
    c = np.array([[1.0]], order='F', dtype=float)
    d = np.array([[1.0]], order='F', dtype=float)
    e = np.array([[4.0]], order='F', dtype=float)
    f = np.array([[2.0]], order='F', dtype=float)

    c_orig = c.copy()
    f_orig = f.copy()

    r, l_mat, scale, info = mb04rt(m, n, pmax, a, b, c, d, e, f)

    assert info == 0
    assert scale > 0

    rhs1 = a @ r - l_mat @ b
    rhs2 = d @ r - l_mat @ e

    np.testing.assert_allclose(rhs1, scale * c_orig, rtol=1e-14)
    np.testing.assert_allclose(rhs2, scale * f_orig, rtol=1e-14)


def test_mb04rt_2x2_diagonal():
    """
    Test 2x2 diagonal system.

    Eigenvalues of (A,D): 1/0.5=2, 3/1=3
    Eigenvalues of (B,E): 4/1=4, 5/2=2.5
    Distinct eigenvalues ensure well-posed system.

    Random seed: 123 (for reproducibility)
    """
    m, n = 2, 2
    pmax = 1e10

    a = np.diag([1.0, 3.0]).astype(float, order='F')
    b = np.diag([4.0, 5.0]).astype(float, order='F')
    d = np.diag([0.5, 1.0]).astype(float, order='F')
    e = np.diag([1.0, 2.0]).astype(float, order='F')

    c = np.array([[1.0, 2.0], [3.0, 4.0]], order='F', dtype=float)
    f = np.array([[0.5, 1.0], [1.5, 2.0]], order='F', dtype=float)

    c_orig = c.copy()
    f_orig = f.copy()

    r, l_mat, scale, info = mb04rt(m, n, pmax, a, b, c, d, e, f)

    assert info == 0
    assert scale > 0

    rhs1 = a @ r - l_mat @ b
    rhs2 = d @ r - l_mat @ e

    np.testing.assert_allclose(rhs1, scale * c_orig, rtol=1e-13, atol=1e-14)
    np.testing.assert_allclose(rhs2, scale * f_orig, rtol=1e-13, atol=1e-14)


def test_mb04rt_2x2_block_a():
    """
    Test with 2x2 block in A (quasi-triangular, complex eigenvalues).

    Random seed: 456 (for reproducibility)
    """
    m, n = 2, 1
    pmax = 1e10

    a = np.array([[1.0, 2.0],
                  [-0.5, 1.0]], order='F', dtype=float)
    b = np.array([[3.0]], order='F', dtype=float)

    d = np.array([[1.0, 0.5],
                  [0.0, 1.0]], order='F', dtype=float)
    e = np.array([[2.0]], order='F', dtype=float)

    c = np.array([[1.0], [2.0]], order='F', dtype=float)
    f = np.array([[0.5], [1.0]], order='F', dtype=float)

    c_orig = c.copy()
    f_orig = f.copy()

    r, l_mat, scale, info = mb04rt(m, n, pmax, a, b, c, d, e, f)

    assert info == 0
    assert scale > 0

    rhs1 = a @ r - l_mat @ b
    rhs2 = d @ r - l_mat @ e

    np.testing.assert_allclose(rhs1, scale * c_orig, rtol=1e-13, atol=1e-14)
    np.testing.assert_allclose(rhs2, scale * f_orig, rtol=1e-13, atol=1e-14)


def test_mb04rt_larger_system_blocked():
    """
    Test larger system that should trigger blocked algorithm.

    Uses 8x6 system with mixed 1x1 and 2x2 blocks to exercise block partitioning.
    Block detection: 2x2 block when a[i, i-1] != 0.

    Random seed: 555 (for reproducibility)
    """
    np.random.seed(555)
    m, n = 8, 6
    pmax = 1e10

    a = np.zeros((m, m), order='F', dtype=float)
    a[0, 0] = 1.0
    a[0, 1] = 2.0
    a[1, 0] = -0.5
    a[1, 1] = 1.0
    a[2, 2] = 2.5
    a[2, 3] = 0.1
    a[3, 3] = 3.0
    a[4, 4] = 1.5
    a[4, 5] = 1.2
    a[5, 4] = -0.4
    a[5, 5] = 1.5
    a[6, 6] = 4.0
    a[7, 7] = 5.0
    for i in range(m):
        for j in range(i + 2, m):
            a[i, j] = np.random.randn() * 0.1

    b = np.zeros((n, n), order='F', dtype=float)
    b[0, 0] = 2.0
    b[1, 1] = 1.5
    b[1, 2] = 0.8
    b[2, 1] = -0.3
    b[2, 2] = 1.5
    b[3, 3] = 3.5
    b[4, 4] = 2.5
    b[5, 5] = 4.0
    for i in range(n):
        for j in range(i + 2, n):
            b[i, j] = np.random.randn() * 0.1

    d = np.eye(m, order='F', dtype=float)
    for i in range(m):
        d[i, i] = 0.5 + i * 0.2
    for i in range(m):
        for j in range(i + 1, m):
            d[i, j] = np.random.randn() * 0.05

    e = np.eye(n, order='F', dtype=float)
    for i in range(n):
        e[i, i] = 1.0 + i * 0.3
    for i in range(n):
        for j in range(i + 1, n):
            e[i, j] = np.random.randn() * 0.05

    c = np.random.randn(m, n).astype(float, order='F')
    f = np.random.randn(m, n).astype(float, order='F')

    c_orig = c.copy()
    f_orig = f.copy()

    r, l_mat, scale, info = mb04rt(m, n, pmax, a, b, c, d, e, f)

    assert info == 0
    assert scale > 0

    rhs1 = a @ r - l_mat @ b
    rhs2 = d @ r - l_mat @ e

    np.testing.assert_allclose(rhs1, scale * c_orig, rtol=1e-12, atol=1e-13)
    np.testing.assert_allclose(rhs2, scale * f_orig, rtol=1e-12, atol=1e-13)


def test_mb04rt_pmax_exceeded():
    """
    Test that INFO=1 when solution exceeds PMAX.

    Random seed: 111 (for reproducibility)
    """
    m, n = 1, 1
    pmax = 0.001

    a = np.array([[0.001]], order='F', dtype=float)
    b = np.array([[0.001]], order='F', dtype=float)
    c = np.array([[100.0]], order='F', dtype=float)
    d = np.array([[0.001]], order='F', dtype=float)
    e = np.array([[0.002]], order='F', dtype=float)
    f = np.array([[100.0]], order='F', dtype=float)

    r, l_mat, scale, info = mb04rt(m, n, pmax, a, b, c, d, e, f)

    assert info == 1


def test_mb04rt_zero_m():
    """
    Test quick return for M=0.
    """
    pmax = 1e10

    a = np.array([], order='F', dtype=float).reshape(0, 0)
    b = np.array([[1.0]], order='F', dtype=float)
    c = np.array([], order='F', dtype=float).reshape(0, 1)
    d = np.array([], order='F', dtype=float).reshape(0, 0)
    e = np.array([[1.0]], order='F', dtype=float)
    f = np.array([], order='F', dtype=float).reshape(0, 1)

    r, l_mat, scale, info = mb04rt(0, 1, pmax, a, b, c, d, e, f)
    assert info == 0
    assert scale == 1.0


def test_mb04rt_zero_n():
    """
    Test quick return for N=0.
    """
    pmax = 1e10

    a = np.array([[1.0]], order='F', dtype=float)
    b = np.array([], order='F', dtype=float).reshape(0, 0)
    c = np.array([], order='F', dtype=float).reshape(1, 0)
    d = np.array([[1.0]], order='F', dtype=float)
    e = np.array([], order='F', dtype=float).reshape(0, 0)
    f = np.array([], order='F', dtype=float).reshape(1, 0)

    r, l_mat, scale, info = mb04rt(1, 0, pmax, a, b, c, d, e, f)
    assert info == 0
    assert scale == 1.0


def test_mb04rt_residual_property():
    """
    Mathematical property test: verify residual equations hold.

    For any valid solution:
        A * R - L * B = scale * C
        D * R - L * E = scale * F

    Random seed: 777 (for reproducibility)
    """
    np.random.seed(777)
    m, n = 5, 4
    pmax = 1e10

    a = np.zeros((m, m), order='F', dtype=float)
    a[0, 0] = 2.0
    a[1, 1] = 1.5
    a[1, 2] = 0.3
    a[2, 2] = 3.0
    a[3, 3] = 1.0
    a[3, 4] = 0.8
    a[4, 3] = -0.2
    a[4, 4] = 1.0
    for i in range(m):
        for j in range(i + 2, m):
            if not (i == 3 and j == 4):
                a[i, j] = np.random.randn() * 0.1

    b = np.zeros((n, n), order='F', dtype=float)
    b[0, 0] = 1.0
    b[0, 1] = 2.0
    b[1, 0] = -0.3
    b[1, 1] = 1.0
    b[2, 2] = 2.0
    b[3, 3] = 3.0
    for i in range(n):
        for j in range(i + 2, n):
            if not (i == 0 and j == 1):
                b[i, j] = np.random.randn() * 0.1

    d = np.eye(m, order='F', dtype=float)
    for i in range(m):
        d[i, i] = 1.0 + i * 0.2
    for i in range(m):
        for j in range(i + 1, m):
            d[i, j] = np.random.randn() * 0.02

    e = np.eye(n, order='F', dtype=float)
    for i in range(n):
        e[i, i] = 2.0 + i * 0.3
    for i in range(n):
        for j in range(i + 1, n):
            e[i, j] = np.random.randn() * 0.02

    c = np.random.randn(m, n).astype(float, order='F')
    f = np.random.randn(m, n).astype(float, order='F')

    c_orig = c.copy()
    f_orig = f.copy()

    r, l_mat, scale, info = mb04rt(m, n, pmax, a, b, c, d, e, f)

    assert info == 0
    assert 0 < scale <= 1.0

    residual1 = a @ r - l_mat @ b - scale * c_orig
    residual2 = d @ r - l_mat @ e - scale * f_orig

    np.testing.assert_allclose(residual1, 0.0, atol=1e-12)
    np.testing.assert_allclose(residual2, 0.0, atol=1e-12)


def test_mb04rt_matches_mb04rs():
    """
    Verify MB04RT produces same result as MB04RS (unblocked).

    For small systems where blocking shouldn't change result.

    Random seed: 888 (for reproducibility)
    """
    from slicot import mb04rs

    np.random.seed(888)
    m, n = 3, 2
    pmax = 1e10

    a = np.array([
        [2.0, 1.0, 0.5],
        [0.0, 1.5, 0.3],
        [0.0, 0.0, 3.0]
    ], order='F', dtype=float)

    b = np.array([
        [1.0, 2.0],
        [-0.3, 1.0]
    ], order='F', dtype=float)

    d = np.array([
        [1.0, 0.5, 0.2],
        [0.0, 1.0, 0.1],
        [0.0, 0.0, 1.5]
    ], order='F', dtype=float)

    e = np.array([
        [2.0, 0.3],
        [0.0, 2.0]
    ], order='F', dtype=float)

    c = np.random.randn(m, n).astype(float, order='F')
    f = np.random.randn(m, n).astype(float, order='F')

    c1 = c.copy()
    f1 = f.copy()
    c2 = c.copy()
    f2 = f.copy()

    r1, l1, scale1, info1 = mb04rs(m, n, pmax, a, b, c1, d, e, f1)
    r2, l2, scale2, info2 = mb04rt(m, n, pmax, a, b, c2, d, e, f2)

    assert info1 == 0
    assert info2 == 0

    np.testing.assert_allclose(scale1, scale2, rtol=1e-14)
    np.testing.assert_allclose(r1, r2, rtol=1e-13, atol=1e-14)
    np.testing.assert_allclose(l1, l2, rtol=1e-13, atol=1e-14)


def test_mb04rt_block_boundary_detection():
    """
    Test that 2x2 blocks are detected at block boundaries.

    Uses a 4x4 system where a 2x2 block spans what would otherwise be
    the block boundary. Must not split 2x2 blocks.

    Random seed: 999 (for reproducibility)
    """
    m, n = 4, 4
    pmax = 1e10

    a = np.array([
        [1.0, 0.5, 0.0, 0.0],
        [0.0, 2.0, 1.0, 0.0],
        [0.0, -0.5, 2.0, 0.0],
        [0.0, 0.0, 0.0, 3.0]
    ], order='F', dtype=float)

    b = np.array([
        [1.5, 0.0, 0.0, 0.0],
        [0.0, 1.0, 2.0, 0.0],
        [0.0, -0.3, 1.0, 0.0],
        [0.0, 0.0, 0.0, 2.5]
    ], order='F', dtype=float)

    d = np.diag([1.0, 1.0, 1.0, 1.0]).astype(float, order='F')
    e = np.diag([1.0, 1.0, 1.0, 1.0]).astype(float, order='F')

    c = np.array([
        [1.0, 2.0, 3.0, 4.0],
        [5.0, 6.0, 7.0, 8.0],
        [9.0, 10.0, 11.0, 12.0],
        [13.0, 14.0, 15.0, 16.0]
    ], order='F', dtype=float)

    f = np.array([
        [0.1, 0.2, 0.3, 0.4],
        [0.5, 0.6, 0.7, 0.8],
        [0.9, 1.0, 1.1, 1.2],
        [1.3, 1.4, 1.5, 1.6]
    ], order='F', dtype=float)

    c_orig = c.copy()
    f_orig = f.copy()

    r, l_mat, scale, info = mb04rt(m, n, pmax, a, b, c, d, e, f)

    assert info == 0
    assert scale > 0

    rhs1 = a @ r - l_mat @ b
    rhs2 = d @ r - l_mat @ e

    np.testing.assert_allclose(rhs1, scale * c_orig, rtol=1e-12, atol=1e-13)
    np.testing.assert_allclose(rhs2, scale * f_orig, rtol=1e-12, atol=1e-13)



"""
Tests for TG01CD - QR-coordinate form reduction of descriptor system.

TG01CD reduces the descriptor system pair (A-lambda E, B) to QR-coordinate form
by computing an orthogonal transformation matrix Q such that Q'*E is upper
trapezoidal.

Test data sources:
- SLICOT HTML documentation TG01CD.html example
- Mathematical properties of QR transformation
"""

import numpy as np
import pytest

from slicot import tg01cd


def test_tg01cd_html_example():
    """
    Test using HTML documentation example data.

    Validates numerical correctness against published example.
    """
    l, n, m = 4, 4, 2

    # Input from HTML: row-wise (( A(I,J), J = 1,N ), I = 1,L)
    a = np.array([
        [-1.0,  0.0,  0.0,  3.0],
        [ 0.0,  0.0,  1.0,  2.0],
        [ 1.0,  1.0,  0.0,  4.0],
        [ 0.0,  0.0,  0.0,  0.0]
    ], order='F', dtype=float)

    e = np.array([
        [1.0, 2.0, 0.0, 0.0],
        [0.0, 1.0, 0.0, 1.0],
        [3.0, 9.0, 6.0, 3.0],
        [0.0, 0.0, 2.0, 0.0]
    ], order='F', dtype=float)

    b = np.array([
        [1.0, 0.0],
        [0.0, 0.0],
        [0.0, 1.0],
        [1.0, 1.0]
    ], order='F', dtype=float)

    a_out, e_out, b_out, q_out, info = tg01cd('I', a, e, b)

    assert info == 0

    # Expected from HTML docs
    a_expected = np.array([
        [-0.6325, -0.9487,  0.0000, -4.7434],
        [-0.8706, -0.2176, -0.7255, -0.3627],
        [-0.5203, -0.1301,  0.3902,  1.4307],
        [-0.7559, -0.1890,  0.5669,  2.0788]
    ], order='F', dtype=float)

    e_expected = np.array([
        [-3.1623, -9.1706, -5.6921, -2.8460],
        [ 0.0000, -1.3784, -1.3059, -1.3784],
        [ 0.0000,  0.0000, -2.4279,  0.0000],
        [ 0.0000,  0.0000,  0.0000,  0.0000]
    ], order='F', dtype=float)

    b_expected = np.array([
        [-0.3162, -0.9487],
        [ 0.6529, -0.2176],
        [-0.4336, -0.9538],
        [ 1.1339,  0.3780]
    ], order='F', dtype=float)

    q_expected = np.array([
        [-0.3162,  0.6529,  0.3902,  0.5669],
        [ 0.0000, -0.7255,  0.3902,  0.5669],
        [-0.9487, -0.2176, -0.1301, -0.1890],
        [ 0.0000,  0.0000, -0.8238,  0.5669]
    ], order='F', dtype=float)

    # HTML shows 4 decimal places, use rtol=1e-3
    np.testing.assert_allclose(a_out, a_expected, rtol=1e-3, atol=1e-4)
    np.testing.assert_allclose(e_out, e_expected, rtol=1e-3, atol=1e-4)
    np.testing.assert_allclose(b_out, b_expected, rtol=1e-3, atol=1e-4)
    np.testing.assert_allclose(q_out, q_expected, rtol=1e-3, atol=1e-4)


def test_tg01cd_upper_trapezoidal():
    """
    Test that Q'*E is upper trapezoidal after transformation.

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    l, n, m = 5, 4, 2

    a = np.random.randn(l, n).astype(float, order='F')
    e = np.random.randn(l, n).astype(float, order='F')
    b = np.random.randn(l, m).astype(float, order='F')

    a_out, e_out, b_out, q_out, info = tg01cd('I', a, e, b)

    assert info == 0

    # E should be upper trapezoidal (zeros below diagonal in first min(l,n) columns)
    ln = min(l, n)
    for i in range(ln + 1, l):
        for j in range(ln):
            assert abs(e_out[i, j]) < 1e-14, f"E[{i},{j}] = {e_out[i, j]} should be zero"


def test_tg01cd_orthogonality():
    """
    Test that Q is orthogonal: Q'*Q = I.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    l, n, m = 4, 4, 2

    a = np.random.randn(l, n).astype(float, order='F')
    e = np.random.randn(l, n).astype(float, order='F')
    b = np.random.randn(l, m).astype(float, order='F')

    a_out, e_out, b_out, q_out, info = tg01cd('I', a, e, b)

    assert info == 0

    # Q should be orthogonal
    qtq = q_out.T @ q_out
    np.testing.assert_allclose(qtq, np.eye(l), rtol=1e-14, atol=1e-14)


def test_tg01cd_transformation_consistency():
    """
    Test that transformations are consistent: A_out = Q'*A, E_out = Q'*E, B_out = Q'*B.

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    l, n, m = 4, 4, 2

    a = np.random.randn(l, n).astype(float, order='F')
    e = np.random.randn(l, n).astype(float, order='F')
    b = np.random.randn(l, m).astype(float, order='F')

    a_orig = a.copy()
    e_orig = e.copy()
    b_orig = b.copy()

    a_out, e_out, b_out, q_out, info = tg01cd('I', a, e, b)

    assert info == 0

    # A_out = Q' * A
    a_check = q_out.T @ a_orig
    np.testing.assert_allclose(a_out, a_check, rtol=1e-13, atol=1e-14)

    # E_out = Q' * E
    e_check = q_out.T @ e_orig
    np.testing.assert_allclose(e_out, e_check, rtol=1e-13, atol=1e-14)

    # B_out = Q' * B
    b_check = q_out.T @ b_orig
    np.testing.assert_allclose(b_out, b_check, rtol=1e-13, atol=1e-14)


def test_tg01cd_compq_n():
    """
    Test with COMPQ='N' (do not compute Q).

    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)
    l, n, m = 3, 3, 1

    a = np.random.randn(l, n).astype(float, order='F')
    e = np.random.randn(l, n).astype(float, order='F')
    b = np.random.randn(l, m).astype(float, order='F')

    a_out, e_out, b_out, q_out, info = tg01cd('N', a, e, b)

    assert info == 0

    # E should still be upper trapezoidal
    for i in range(1, l):
        for j in range(min(i, n)):
            assert abs(e_out[i, j]) < 1e-14


def test_tg01cd_compq_u():
    """
    Test with COMPQ='U' (update existing Q).

    Random seed: 111 (for reproducibility)
    """
    np.random.seed(111)
    l, n, m = 4, 4, 2

    a = np.random.randn(l, n).astype(float, order='F')
    e = np.random.randn(l, n).astype(float, order='F')
    b = np.random.randn(l, m).astype(float, order='F')

    # Create initial orthogonal Q1 from QR of random matrix
    q1_temp = np.random.randn(l, l).astype(float, order='F')
    q1, _ = np.linalg.qr(q1_temp)
    q1 = np.asfortranarray(q1)

    a_out, e_out, b_out, q_out, info = tg01cd('U', a, e, b, q1.copy())

    assert info == 0

    # Q_out should still be orthogonal
    qtq = q_out.T @ q_out
    np.testing.assert_allclose(qtq, np.eye(l), rtol=1e-14, atol=1e-14)


def test_tg01cd_wide_matrix():
    """
    Test with L < N (wide matrix, more columns than rows).

    Random seed: 222 (for reproducibility)
    """
    np.random.seed(222)
    l, n, m = 3, 5, 2

    a = np.random.randn(l, n).astype(float, order='F')
    e = np.random.randn(l, n).astype(float, order='F')
    b = np.random.randn(l, m).astype(float, order='F')

    a_out, e_out, b_out, q_out, info = tg01cd('I', a, e, b)

    assert info == 0

    # E should be upper trapezoidal: E11 (L x L upper triangular), E12 (L x (N-L))
    for i in range(1, l):
        for j in range(i):
            assert abs(e_out[i, j]) < 1e-14, f"E[{i},{j}] = {e_out[i, j]} should be zero"


def test_tg01cd_tall_matrix():
    """
    Test with L > N (tall matrix, more rows than columns).

    Random seed: 333 (for reproducibility)
    """
    np.random.seed(333)
    l, n, m = 5, 3, 2

    a = np.random.randn(l, n).astype(float, order='F')
    e = np.random.randn(l, n).astype(float, order='F')
    b = np.random.randn(l, m).astype(float, order='F')

    a_out, e_out, b_out, q_out, info = tg01cd('I', a, e, b)

    assert info == 0

    # E should be upper trapezoidal with zeros in rows N+1:L
    for i in range(n, l):
        for j in range(n):
            assert abs(e_out[i, j]) < 1e-14, f"E[{i},{j}] = {e_out[i, j]} should be zero"


def test_tg01cd_zero_m():
    """
    Test with M=0 (no B matrix columns).

    Random seed: 444 (for reproducibility)
    """
    np.random.seed(444)
    l, n, m = 4, 4, 0

    a = np.random.randn(l, n).astype(float, order='F')
    e = np.random.randn(l, n).astype(float, order='F')
    b = np.empty((l, 0), order='F', dtype=float)

    a_out, e_out, b_out, q_out, info = tg01cd('I', a, e, b)

    assert info == 0
    assert b_out.shape == (l, 0)


def test_tg01cd_quick_return():
    """
    Test quick return for L=0 or N=0.
    """
    # L=0
    a = np.empty((0, 4), order='F', dtype=float)
    e = np.empty((0, 4), order='F', dtype=float)
    b = np.empty((0, 2), order='F', dtype=float)

    a_out, e_out, b_out, q_out, info = tg01cd('I', a, e, b)
    assert info == 0

    # N=0
    a = np.empty((4, 0), order='F', dtype=float)
    e = np.empty((4, 0), order='F', dtype=float)
    b = np.random.randn(4, 2).astype(float, order='F')

    a_out, e_out, b_out, q_out, info = tg01cd('I', a, e, b)
    assert info == 0


def test_tg01cd_invalid_compq():
    """
    Test error handling for invalid COMPQ parameter.
    """
    l, n, m = 3, 3, 1

    a = np.zeros((l, n), order='F', dtype=float)
    e = np.zeros((l, n), order='F', dtype=float)
    b = np.zeros((l, m), order='F', dtype=float)

    a_out, e_out, b_out, q_out, info = tg01cd('X', a, e, b)
    assert info == -1



"""
Tests for MB04RZ - Block-diagonalization of generalized complex Schur form.

Reduces a complex matrix pair (A,B) in generalized complex Schur form to
block-diagonal form using well-conditioned non-unitary equivalence
transformations.

The transformations are bounded by PMAX and optionally accumulated in X and Y.
Optionally reorders diagonal elements so clustered eigenvalues are grouped.
"""

import numpy as np
import pytest
from slicot import mb04rz


def test_mb04rz_basic_2x2_diagonal():
    """
    Test 2x2 diagonal complex system - simplest case with distinct eigenvalues.

    Both A and B are diagonal (trivially in upper triangular Schur form).
    Eigenvalues: A11/B11 = (1+1j)/1 = 1+1j, A22/B22 = (4+2j)/2 = 2+1j.
    Already block-diagonal, so should return with minimal blocks.

    Random seed: 42 (for reproducibility)
    """
    n = 2
    pmax = 1e10
    tol = 0.0

    a = np.array([[1.0+1.0j, 0.0],
                  [0.0, 4.0+2.0j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.0],
                  [0.0, 2.0+0.0j]], order='F', dtype=complex)
    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'U', 'N', n, pmax, a, b, x, y, tol)

    assert info == 0
    assert nblcks >= 1
    assert np.sum(blsize[:nblcks]) == n

    for i in range(n):
        assert beta[i].real >= 0
        assert beta[i].imag == pytest.approx(0.0, abs=1e-14)

    assert alpha[0] == pytest.approx(1.0+1.0j, rel=1e-14)
    assert alpha[1] == pytest.approx(4.0+2.0j, rel=1e-14)
    assert beta[0].real == pytest.approx(1.0, rel=1e-14)
    assert beta[1].real == pytest.approx(2.0, rel=1e-14)


def test_mb04rz_3x3_upper_triangular():
    """
    Test 3x3 upper triangular complex system.

    A and B are upper triangular (complex Schur form).

    Random seed: 123 (for reproducibility)
    """
    n = 3
    pmax = 1e10
    tol = 0.0

    a = np.array([[1.0+0.5j, 2.0+1.0j, 0.5+0.2j],
                  [0.0, 1.5-0.5j, 0.3+0.1j],
                  [0.0, 0.0, 3.0+1.0j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.5+0.1j, 0.2+0.05j],
                  [0.0, 1.0+0.0j, 0.1+0.02j],
                  [0.0, 0.0, 2.0+0.0j]], order='F', dtype=complex)
    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'U', 'N', n, pmax, a, b, x, y, tol)

    assert info == 0
    assert np.sum(blsize[:nblcks]) == n

    for i in range(n):
        assert beta[i].real >= 0


def test_mb04rz_4x4_sorted_clustering():
    """
    Test 4x4 system with eigenvalue clustering (SORT='S').

    Create a system where eigenvalues cluster and should be grouped.

    Random seed: 456 (for reproducibility)
    """
    n = 4
    pmax = 1e10
    tol = 0.5

    a = np.array([[1.0+0.0j, 0.3+0.1j, 0.1+0.05j, 0.05+0.02j],
                  [0.0, 1.1+0.0j, 0.2+0.1j, 0.1+0.05j],
                  [0.0, 0.0, 3.0+1.0j, 0.5+0.2j],
                  [0.0, 0.0, 0.0, 3.2+1.0j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.2+0.05j, 0.1+0.02j, 0.05+0.01j],
                  [0.0, 1.0+0.0j, 0.1+0.02j, 0.05+0.01j],
                  [0.0, 0.0, 1.0+0.0j, 0.3+0.1j],
                  [0.0, 0.0, 0.0, 1.0+0.0j]], order='F', dtype=complex)
    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'U', 'S', n, pmax, a, b, x, y, tol)

    assert info == 0
    assert nblcks >= 1
    assert np.sum(blsize[:nblcks]) == n


def test_mb04rz_no_accumulation():
    """
    Test without accumulating transformations (JOBX='N', JOBY='N').

    Random seed: 789 (for reproducibility)
    """
    n = 3
    pmax = 1e10
    tol = 0.0

    a = np.array([[2.0+1.0j, 1.0+0.5j, 0.5+0.2j],
                  [0.0, 1.5+0.3j, 0.3+0.1j],
                  [0.0, 0.0, 3.0+0.5j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.3+0.1j, 0.1+0.05j],
                  [0.0, 1.0+0.0j, 0.2+0.05j],
                  [0.0, 0.0, 2.0+0.0j]], order='F', dtype=complex)
    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'N', 'N', 'N', n, pmax, a, b, x, y, tol)

    assert info == 0

    for i in range(n):
        assert alpha[i] != 0 or beta[i] != 0


def test_mb04rz_quick_return_n0():
    """
    Test quick return for N=0.
    """
    pmax = 1e10
    tol = 0.0

    a = np.array([], order='F', dtype=complex).reshape(0, 0)
    b = np.array([], order='F', dtype=complex).reshape(0, 0)
    x = np.array([], order='F', dtype=complex).reshape(0, 0)
    y = np.array([], order='F', dtype=complex).reshape(0, 0)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'U', 'N', 0, pmax, a, b, x, y, tol)

    assert info == 0
    assert nblcks == 0


def test_mb04rz_quick_return_n1():
    """
    Test quick return for N=1.
    """
    n = 1
    pmax = 1e10
    tol = 0.0

    a = np.array([[5.0+2.0j]], order='F', dtype=complex)
    b = np.array([[2.0+0.0j]], order='F', dtype=complex)
    x = np.array([[1.0+0.0j]], order='F', dtype=complex)
    y = np.array([[1.0+0.0j]], order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'U', 'N', n, pmax, a, b, x, y, tol)

    assert info == 0
    assert nblcks == 1
    assert blsize[0] == 1
    assert alpha[0] == pytest.approx(5.0+2.0j, rel=1e-14)
    assert beta[0].real == pytest.approx(2.0, rel=1e-14)


def test_mb04rz_closest_neighbour():
    """
    Test SORT='C' (closest-neighbour strategy without reordering).

    Random seed: 555 (for reproducibility)
    """
    n = 4
    pmax = 1e10
    tol = 0.0

    a = np.array([[2.0+0.5j, 1.0+0.3j, 0.5+0.1j, 0.1+0.05j],
                  [0.0, 1.5+0.2j, 0.3+0.1j, 0.2+0.05j],
                  [0.0, 0.0, 3.0+1.0j, 0.4+0.2j],
                  [0.0, 0.0, 0.0, 2.5+0.8j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.2+0.05j, 0.1+0.02j, 0.05+0.01j],
                  [0.0, 1.0+0.0j, 0.15+0.03j, 0.08+0.02j],
                  [0.0, 0.0, 1.5+0.0j, 0.2+0.05j],
                  [0.0, 0.0, 0.0, 1.2+0.0j]], order='F', dtype=complex)
    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'U', 'C', n, pmax, a, b, x, y, tol)

    assert info == 0
    assert nblcks >= 1


def test_mb04rz_both_strategies():
    """
    Test SORT='B' (reordering with closest-neighbour strategy).

    Random seed: 666 (for reproducibility)
    """
    n = 4
    pmax = 1e10
    tol = 0.3

    a = np.array([[1.0+0.2j, 0.5+0.1j, 0.2+0.05j, 0.1+0.02j],
                  [0.0, 1.2+0.3j, 0.3+0.1j, 0.15+0.05j],
                  [0.0, 0.0, 2.0+0.5j, 0.4+0.1j],
                  [0.0, 0.0, 0.0, 2.3+0.6j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.1+0.02j, 0.05+0.01j, 0.02+0.005j],
                  [0.0, 1.0+0.0j, 0.1+0.02j, 0.05+0.01j],
                  [0.0, 0.0, 1.0+0.0j, 0.2+0.05j],
                  [0.0, 0.0, 0.0, 1.0+0.0j]], order='F', dtype=complex)
    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'U', 'B', n, pmax, a, b, x, y, tol)

    assert info == 0
    assert nblcks >= 1


def test_mb04rz_eigenvalue_preservation():
    """
    Mathematical property: eigenvalues should be preserved.

    The block-diagonalization should not change the eigenvalues of the
    pencil (A,B). All alpha/beta should match original diagonal.

    Random seed: 777 (for reproducibility)
    """
    np.random.seed(777)
    n = 4
    pmax = 1e10
    tol = 0.0

    a = np.array([[2.0+0.5j, 0.5+0.1j, 0.2+0.05j, 0.1+0.02j],
                  [0.0, 1.5+0.3j, 0.3+0.1j, 0.15+0.05j],
                  [0.0, 0.0, 3.0+1.0j, 0.4+0.1j],
                  [0.0, 0.0, 0.0, 2.5+0.8j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.2+0.05j, 0.1+0.02j, 0.05+0.01j],
                  [0.0, 1.0+0.0j, 0.15+0.03j, 0.08+0.02j],
                  [0.0, 0.0, 1.5+0.0j, 0.2+0.05j],
                  [0.0, 0.0, 0.0, 1.2+0.0j]], order='F', dtype=complex)

    eig_before = []
    for i in range(n):
        if b[i, i].real > 1e-15:
            eig_before.append(a[i, i] / b[i, i].real)
        else:
            eig_before.append(np.inf)

    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'U', 'N', n, pmax, a, b, x, y, tol)

    assert info == 0

    eig_after = []
    for i in range(n):
        if beta[i].real > 1e-15:
            eig_after.append(alpha[i] / beta[i].real)
        else:
            eig_after.append(np.inf)

    eig_before = np.array(eig_before)
    eig_after = np.array(eig_after)

    finite_before = eig_before[np.isfinite(eig_before)]
    finite_after = eig_after[np.isfinite(eig_after)]

    if len(finite_before) == len(finite_after):
        finite_before_sorted = np.sort(np.real(finite_before))
        finite_after_sorted = np.sort(np.real(finite_after))
        np.testing.assert_allclose(finite_before_sorted, finite_after_sorted, rtol=1e-10)


def test_mb04rz_block_structure():
    """
    Mathematical property: output A should be block upper triangular.

    After block-diagonalization, A should have zeros below each diagonal block.

    Random seed: 999 (for reproducibility)
    """
    n = 4
    pmax = 1e10
    tol = 0.0

    a = np.array([[2.0+0.5j, 0.5+0.1j, 0.2+0.05j, 0.1+0.02j],
                  [0.0, 1.5+0.3j, 0.3+0.1j, 0.15+0.05j],
                  [0.0, 0.0, 3.0+1.0j, 0.4+0.1j],
                  [0.0, 0.0, 0.0, 2.5+0.8j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.2+0.05j, 0.1+0.02j, 0.05+0.01j],
                  [0.0, 1.0+0.0j, 0.15+0.03j, 0.08+0.02j],
                  [0.0, 0.0, 1.5+0.0j, 0.2+0.05j],
                  [0.0, 0.0, 0.0, 1.2+0.0j]], order='F', dtype=complex)
    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'U', 'N', n, pmax, a, b, x, y, tol)

    assert info == 0

    block_start = 0
    for k in range(nblcks):
        bs = blsize[k]
        block_end = block_start + bs

        if block_end < n:
            off_diag_a = a_out[block_end:, block_start:block_end]
            off_diag_b = b_out[block_end:, block_start:block_end]
            np.testing.assert_allclose(off_diag_a, 0.0, atol=1e-10)
            np.testing.assert_allclose(off_diag_b, 0.0, atol=1e-10)

        block_start = block_end


def test_mb04rz_negative_tol():
    """
    Test with negative TOL (relative tolerance mode).

    Random seed: 111 (for reproducibility)
    """
    n = 4
    pmax = 1e10
    tol = -0.1

    a = np.array([[1.0+0.0j, 0.3+0.1j, 0.1+0.05j, 0.05+0.02j],
                  [0.0, 1.05+0.0j, 0.2+0.1j, 0.1+0.05j],
                  [0.0, 0.0, 3.0+1.0j, 0.5+0.2j],
                  [0.0, 0.0, 0.0, 3.1+1.0j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.2+0.05j, 0.1+0.02j, 0.05+0.01j],
                  [0.0, 1.0+0.0j, 0.1+0.02j, 0.05+0.01j],
                  [0.0, 0.0, 1.0+0.0j, 0.3+0.1j],
                  [0.0, 0.0, 0.0, 1.0+0.0j]], order='F', dtype=complex)
    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'U', 'S', n, pmax, a, b, x, y, tol)

    assert info == 0
    assert nblcks >= 1


def test_mb04rz_error_invalid_jobx():
    """
    Test error handling for invalid JOBX parameter.
    """
    n = 2
    pmax = 1e10
    tol = 0.0

    a = np.array([[1.0+0.0j, 0.5+0.1j],
                  [0.0, 2.0+0.0j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.2+0.05j],
                  [0.0, 1.0+0.0j]], order='F', dtype=complex)
    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'X', 'U', 'N', n, pmax, a, b, x, y, tol)

    assert info == -1


def test_mb04rz_error_invalid_joby():
    """
    Test error handling for invalid JOBY parameter.
    """
    n = 2
    pmax = 1e10
    tol = 0.0

    a = np.array([[1.0+0.0j, 0.5+0.1j],
                  [0.0, 2.0+0.0j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.2+0.05j],
                  [0.0, 1.0+0.0j]], order='F', dtype=complex)
    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'X', 'N', n, pmax, a, b, x, y, tol)

    assert info == -2


def test_mb04rz_error_invalid_sort():
    """
    Test error handling for invalid SORT parameter.
    """
    n = 2
    pmax = 1e10
    tol = 0.0

    a = np.array([[1.0+0.0j, 0.5+0.1j],
                  [0.0, 2.0+0.0j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.2+0.05j],
                  [0.0, 1.0+0.0j]], order='F', dtype=complex)
    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'U', 'X', n, pmax, a, b, x, y, tol)

    assert info == -3


def test_mb04rz_error_invalid_pmax():
    """
    Test error handling for invalid PMAX parameter.
    """
    n = 2
    pmax = 0.5
    tol = 0.0

    a = np.array([[1.0+0.0j, 0.5+0.1j],
                  [0.0, 2.0+0.0j]], order='F', dtype=complex)
    b = np.array([[1.0+0.0j, 0.2+0.05j],
                  [0.0, 1.0+0.0j]], order='F', dtype=complex)
    x = np.eye(n, order='F', dtype=complex)
    y = np.eye(n, order='F', dtype=complex)

    a_out, b_out, x_out, y_out, nblcks, blsize, alpha, beta, info = mb04rz(
        'U', 'U', 'N', n, pmax, a, b, x, y, tol)

    assert info == -5



"""
Tests for MB03GD - Exchange eigenvalues of 2x2 or 4x4 skew-Hamiltonian/Hamiltonian pencil (factored version).

MB03GD computes orthogonal matrix Q and orthogonal symplectic matrix U for a real
regular 2-by-2 or 4-by-4 skew-Hamiltonian/Hamiltonian pencil a J B' J' B - b D with

    B = [[B11, B12], [0, B22]], D = [[D11, D12], [0, -D11']], J = [[0, I], [-I, 0]]

such that J Q' J' D Q and U' B Q keep block triangular form, but eigenvalues are reordered.
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose


def test_mb03gd_n2_basic():
    """
    Test N=2 case with simple 2x2 matrices.

    For N=2, Q and U are Givens rotations computed from B and D.
    Random seed: 42 (for reproducibility)
    """
    from slicot import mb03gd

    n = 2

    b = np.array([
        [2.0, 0.5],
        [0.0, 1.5]
    ], order='F', dtype=float)

    d = np.array([[1.0, 0.3]], order='F', dtype=float)

    macpar = np.array([2.2e-16, 2.2e-308], dtype=float)

    q, u, info = mb03gd(n, b, d, macpar)

    assert info == 0
    assert q.shape == (2, 2)
    assert u.shape == (2, 2)

    np.testing.assert_allclose(q @ q.T, np.eye(2), rtol=1e-14, atol=1e-14)
    np.testing.assert_allclose(u @ u.T, np.eye(2), rtol=1e-14, atol=1e-14)


def test_mb03gd_n2_orthogonality():
    """
    Verify Q and U are orthogonal for N=2 case with various matrices.

    Random seed: 123 (for reproducibility)
    """
    from slicot import mb03gd

    np.random.seed(123)

    n = 2
    macpar = np.array([2.2e-16, 2.2e-308], dtype=float)

    for _ in range(5):
        b11 = np.random.randn() + 1.0
        b12 = np.random.randn()
        b22 = np.random.randn() + 1.0
        b = np.array([
            [b11, b12],
            [0.0, b22]
        ], order='F', dtype=float)

        d11 = np.random.randn()
        d12 = np.random.randn()
        d = np.array([[d11, d12]], order='F', dtype=float)

        q, u, info = mb03gd(n, b, d, macpar)

        assert info == 0
        np.testing.assert_allclose(q @ q.T, np.eye(2), rtol=1e-14, atol=1e-14)
        np.testing.assert_allclose(u @ u.T, np.eye(2), rtol=1e-14, atol=1e-14)


def test_mb03gd_n4_basic():
    """
    Test N=4 case with skew-Hamiltonian/Hamiltonian pencil.

    For N=4:
    - B is 4x4 upper block triangular
    - D is 2x4 (first block row of Hamiltonian matrix)

    Random seed: 456 (for reproducibility)
    """
    from slicot import mb03gd

    n = 4

    b = np.array([
        [2.0, 0.5, 0.3, 0.1],
        [0.0, 1.5, 0.2, 0.4],
        [0.0, 0.0, 1.8, 0.6],
        [0.0, 0.0, 0.0, 2.2]
    ], order='F', dtype=float)

    d = np.array([
        [1.0, 0.3, 0.5, 0.2],
        [0.0, 0.8, 0.1, 0.7]
    ], order='F', dtype=float)

    macpar = np.array([2.2e-16, 2.2e-308], dtype=float)

    q, u, info = mb03gd(n, b, d, macpar)

    assert info == 0
    assert q.shape == (4, 4)
    assert u.shape == (4, 4)

    np.testing.assert_allclose(q @ q.T, np.eye(4), rtol=1e-13, atol=1e-13)
    np.testing.assert_allclose(u @ u.T, np.eye(4), rtol=1e-13, atol=1e-13)


def test_mb03gd_n4_orthogonality():
    """
    Verify Q and U are orthogonal for N=4 case with random matrices.

    Random seed: 789 (for reproducibility)
    """
    from slicot import mb03gd

    np.random.seed(789)

    n = 4
    macpar = np.array([2.2e-16, 2.2e-308], dtype=float)

    for _ in range(3):
        b = np.zeros((4, 4), order='F', dtype=float)
        b[0, 0] = np.random.randn() + 2.0
        b[0, 1] = np.random.randn()
        b[1, 1] = np.random.randn() + 2.0
        b[0, 2] = np.random.randn()
        b[0, 3] = np.random.randn()
        b[1, 2] = np.random.randn()
        b[1, 3] = np.random.randn()
        b[2, 2] = np.random.randn() + 2.0
        b[2, 3] = np.random.randn()
        b[3, 3] = np.random.randn() + 2.0

        d = np.zeros((2, 4), order='F', dtype=float)
        d[0, 0] = np.random.randn()
        d[0, 1] = np.random.randn()
        d[1, 1] = np.random.randn()
        d[0, 2] = np.random.randn()
        d[0, 3] = np.random.randn()
        d[1, 2] = np.random.randn()
        d[1, 3] = np.random.randn()

        q, u, info = mb03gd(n, b, d, macpar)

        if info == 0:
            np.testing.assert_allclose(q @ q.T, np.eye(4), rtol=1e-13, atol=1e-13)
            np.testing.assert_allclose(u @ u.T, np.eye(4), rtol=1e-13, atol=1e-13)


def test_mb03gd_n4_u_symplectic():
    """
    Verify U is orthogonal symplectic for N=4.

    Orthogonal symplectic matrix U satisfies:
    U^T U = I and U^T J U = J where J = [[0, I], [-I, 0]]

    Random seed: 888 (for reproducibility)
    """
    from slicot import mb03gd

    np.random.seed(888)

    n = 4
    macpar = np.array([2.2e-16, 2.2e-308], dtype=float)

    b = np.array([
        [2.5, 0.3, 0.4, 0.2],
        [0.0, 1.8, 0.1, 0.5],
        [0.0, 0.0, 2.0, 0.3],
        [0.0, 0.0, 0.0, 1.5]
    ], order='F', dtype=float)

    d = np.array([
        [1.2, 0.4, 0.6, 0.3],
        [0.0, 0.9, 0.2, 0.8]
    ], order='F', dtype=float)

    q, u, info = mb03gd(n, b, d, macpar)

    assert info == 0

    np.testing.assert_allclose(u @ u.T, np.eye(4), rtol=1e-13, atol=1e-13)

    J = np.array([
        [0, 0, 1, 0],
        [0, 0, 0, 1],
        [-1, 0, 0, 0],
        [0, -1, 0, 0]
    ], dtype=float)

    utju = u.T @ J @ u
    np.testing.assert_allclose(utju, J, rtol=1e-13, atol=1e-13)


def test_mb03gd_singular_b11():
    """
    Test INFO=1 when B11 is numerically singular.

    Random seed: 999 (for reproducibility)
    """
    from slicot import mb03gd

    n = 4

    b = np.array([
        [1e-320, 1e-320, 0.3, 0.1],
        [0.0, 1e-320, 0.2, 0.4],
        [0.0, 0.0, 1.8, 0.6],
        [0.0, 0.0, 0.0, 2.2]
    ], order='F', dtype=float)

    d = np.array([
        [1.0, 0.3, 0.5, 0.2],
        [0.0, 0.8, 0.1, 0.7]
    ], order='F', dtype=float)

    macpar = np.array([2.2e-16, 2.2e-308], dtype=float)

    q, u, info = mb03gd(n, b, d, macpar)

    assert info == 1


def test_mb03gd_n2_determinant():
    """
    Verify det(Q) = +/-1 and det(U) = 1 for N=2.

    Random seed: 1111 (for reproducibility)
    """
    from slicot import mb03gd

    np.random.seed(1111)

    n = 2
    macpar = np.array([2.2e-16, 2.2e-308], dtype=float)

    b = np.array([
        [1.5, 0.7],
        [0.0, 2.0]
    ], order='F', dtype=float)

    d = np.array([[0.8, 0.4]], order='F', dtype=float)

    q, u, info = mb03gd(n, b, d, macpar)

    assert info == 0

    det_q = np.linalg.det(q)
    det_u = np.linalg.det(u)

    np.testing.assert_allclose(np.abs(det_q), 1.0, rtol=1e-14)
    np.testing.assert_allclose(np.abs(det_u), 1.0, rtol=1e-14)


def test_mb03gd_n4_determinant():
    """
    Verify det(Q) = +/-1 and det(U) = 1 for N=4.

    Random seed: 2222 (for reproducibility)
    """
    from slicot import mb03gd

    np.random.seed(2222)

    n = 4
    macpar = np.array([2.2e-16, 2.2e-308], dtype=float)

    b = np.array([
        [2.0, 0.5, 0.3, 0.1],
        [0.0, 1.5, 0.2, 0.4],
        [0.0, 0.0, 1.8, 0.6],
        [0.0, 0.0, 0.0, 2.2]
    ], order='F', dtype=float)

    d = np.array([
        [1.0, 0.3, 0.5, 0.2],
        [0.0, 0.8, 0.1, 0.7]
    ], order='F', dtype=float)

    q, u, info = mb03gd(n, b, d, macpar)

    assert info == 0

    det_q = np.linalg.det(q)
    det_u = np.linalg.det(u)

    np.testing.assert_allclose(np.abs(det_q), 1.0, rtol=1e-14)
    np.testing.assert_allclose(np.abs(det_u), 1.0, rtol=1e-14)



"""
Tests for SB03SX: Estimate forward error for discrete-time Lyapunov equation.

SB03SX estimates a forward error bound for the solution X of:
    op(A)'*X*op(A) - X = C
where op(A) = A or A' and C is symmetric.
"""

import numpy as np
import pytest
from slicot import sb03sx


def test_sb03sx_basic_notrans():
    """
    Test basic forward error estimation with TRANA='N', LYAPUN='R'.

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n = 4

    t = np.zeros((n, n), order='F', dtype=float)
    for i in range(n):
        t[i, i] = 0.5 + 0.1 * i
        for j in range(i + 1, n):
            t[i, j] = np.random.randn() * 0.1

    xanorm = 1.5

    r = np.random.randn(n, n).astype(float, order='F')
    r = (r + r.T) / 2
    r = np.triu(r)

    u = np.eye(n, order='F', dtype=float)

    ferr, r_out, info = sb03sx('N', 'U', 'R', n, xanorm, t, u, r)

    assert info >= 0
    assert ferr >= 0.0


def test_sb03sx_trans():
    """
    Test with TRANA='T', LYAPUN='R'.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n = 3

    t = np.zeros((n, n), order='F', dtype=float)
    for i in range(n):
        t[i, i] = 0.4 + 0.15 * i
        for j in range(i + 1, n):
            t[i, j] = np.random.randn() * 0.1

    xanorm = 2.0

    r = np.random.randn(n, n).astype(float, order='F')
    r = (r + r.T) / 2
    r = np.tril(r)

    u = np.eye(n, order='F', dtype=float)

    ferr, r_out, info = sb03sx('T', 'L', 'R', n, xanorm, t, u, r)

    assert info >= 0
    assert ferr >= 0.0


def test_sb03sx_original_mode():
    """
    Test with LYAPUN='O' (original mode with orthogonal U).

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    n = 3

    t = np.zeros((n, n), order='F', dtype=float)
    for i in range(n):
        t[i, i] = 0.3 + 0.2 * i
        for j in range(i + 1, n):
            t[i, j] = np.random.randn() * 0.1

    q, _ = np.linalg.qr(np.random.randn(n, n))
    u = q.astype(float, order='F')

    xanorm = 1.0

    r = np.random.randn(n, n).astype(float, order='F')
    r = (r + r.T) / 2
    r = np.triu(r)

    ferr, r_out, info = sb03sx('N', 'U', 'O', n, xanorm, t, u, r)

    assert info >= 0
    assert ferr >= 0.0


def test_sb03sx_n_zero():
    """
    Test quick return for n=0
    """
    n = 0
    xanorm = 1.0

    t = np.array([], order='F', dtype=float).reshape(0, 0)
    u = np.array([], order='F', dtype=float).reshape(0, 0)
    r = np.array([], order='F', dtype=float).reshape(0, 0)

    ferr, r_out, info = sb03sx('N', 'U', 'R', n, xanorm, t, u, r)

    assert info == 0
    assert ferr == 0.0


def test_sb03sx_xanorm_zero():
    """
    Test quick return for xanorm=0
    """
    np.random.seed(789)
    n = 3
    xanorm = 0.0

    t = np.eye(n, order='F', dtype=float)
    u = np.eye(n, order='F', dtype=float)
    r = np.eye(n, order='F', dtype=float)

    ferr, r_out, info = sb03sx('N', 'U', 'R', n, xanorm, t, u, r)

    assert info == 0
    assert ferr == 0.0


def test_sb03sx_invalid_trana():
    """
    Test error handling for invalid TRANA parameter
    """
    n = 2
    t = np.eye(n, order='F', dtype=float)
    u = np.eye(n, order='F', dtype=float)
    r = np.eye(n, order='F', dtype=float)

    ferr, r_out, info = sb03sx('X', 'U', 'R', n, 1.0, t, u, r)

    assert info == -1


def test_sb03sx_invalid_uplo():
    """
    Test error handling for invalid UPLO parameter
    """
    n = 2
    t = np.eye(n, order='F', dtype=float)
    u = np.eye(n, order='F', dtype=float)
    r = np.eye(n, order='F', dtype=float)

    ferr, r_out, info = sb03sx('N', 'X', 'R', n, 1.0, t, u, r)

    assert info == -2


def test_sb03sx_invalid_lyapun():
    """
    Test error handling for invalid LYAPUN parameter
    """
    n = 2
    t = np.eye(n, order='F', dtype=float)
    u = np.eye(n, order='F', dtype=float)
    r = np.eye(n, order='F', dtype=float)

    ferr, r_out, info = sb03sx('N', 'U', 'X', n, 1.0, t, u, r)

    assert info == -3



"""Tests for DG01MD - Discrete Fourier Transform of complex signal."""

import numpy as np
import pytest
from slicot import dg01md


"""Basic functionality tests from SLICOT HTML documentation."""

def test_forward_fft_doc_example():
    """
    Test forward FFT using HTML doc example.

    Input: N=8, INDI='D'
    XR = [-0.1862, 0.3948, 0.6788, 0.1861, 0.7254, 0.5815, 0.4904, -0.9599]
    XI = [0.1288, 0.0671, -0.2417, 0.8875, 0.9380, -0.2682, 0.9312, -0.3116]

    Expected output:
    XR = [1.9109, -1.9419, -1.4070, 2.2886, 1.5059, -2.2271, 0.1470, -1.7660]
    XI = [2.1311, -2.2867, -1.3728, -0.6883, 1.3815, 0.2915, 2.1274, -0.5533]
    """
    xr = np.array([-0.1862, 0.3948, 0.6788, 0.1861, 0.7254, 0.5815, 0.4904, -0.9599], dtype=float)
    xi = np.array([0.1288, 0.0671, -0.2417, 0.8875, 0.9380, -0.2682, 0.9312, -0.3116], dtype=float)

    xr_expected = np.array([1.9109, -1.9419, -1.4070, 2.2886, 1.5059, -2.2271, 0.1470, -1.7660], dtype=float)
    xi_expected = np.array([2.1311, -2.2867, -1.3728, -0.6883, 1.3815, 0.2915, 2.1274, -0.5533], dtype=float)

    xr_out, xi_out, info = dg01md('D', xr, xi)

    assert info == 0
    np.testing.assert_allclose(xr_out, xr_expected, rtol=1e-3, atol=1e-4)
    np.testing.assert_allclose(xi_out, xi_expected, rtol=1e-3, atol=1e-4)


"""Test inverse transform property: FFT then IFFT scales by N."""

def test_forward_inverse_scaling():
    """
    Validate FFT followed by IFFT gives N * original signal.

    This is the key mathematical property documented in SLICOT:
    "a discrete Fourier transform, followed by an inverse discrete
    Fourier transform, will result in a signal which is a factor N
    larger than the original input signal."

    Random seed: 42 (for reproducibility)
    """
    np.random.seed(42)
    n = 8
    xr_orig = np.random.randn(n)
    xi_orig = np.random.randn(n)
    xr = xr_orig.copy()
    xi = xi_orig.copy()

    xr_fft, xi_fft, info1 = dg01md('D', xr, xi)
    assert info1 == 0

    xr_ifft, xi_ifft, info2 = dg01md('I', xr_fft, xi_fft)
    assert info2 == 0

    np.testing.assert_allclose(xr_ifft, n * xr_orig, rtol=1e-14)
    np.testing.assert_allclose(xi_ifft, n * xi_orig, rtol=1e-14)

def test_inverse_forward_scaling():
    """
    Validate IFFT followed by FFT also gives N * original signal.

    Random seed: 123 (for reproducibility)
    """
    np.random.seed(123)
    n = 16
    xr_orig = np.random.randn(n)
    xi_orig = np.random.randn(n)
    xr = xr_orig.copy()
    xi = xi_orig.copy()

    xr_ifft, xi_ifft, info1 = dg01md('I', xr, xi)
    assert info1 == 0

    xr_fft, xi_fft, info2 = dg01md('D', xr_ifft, xi_ifft)
    assert info2 == 0

    np.testing.assert_allclose(xr_fft, n * xr_orig, rtol=1e-14)
    np.testing.assert_allclose(xi_fft, n * xi_orig, rtol=1e-14)


"""Mathematical property tests for DFT."""

def test_linearity_forward():
    """
    Validate DFT linearity: DFT(a*x + b*y) = a*DFT(x) + b*DFT(y).

    Random seed: 456 (for reproducibility)
    """
    np.random.seed(456)
    n = 8
    a, b = 2.5, -1.3

    xr1 = np.random.randn(n)
    xi1 = np.random.randn(n)
    xr2 = np.random.randn(n)
    xi2 = np.random.randn(n)

    xr_sum = a * xr1 + b * xr2
    xi_sum = a * xi1 + b * xi2

    xr1_fft, xi1_fft, _ = dg01md('D', xr1.copy(), xi1.copy())
    xr2_fft, xi2_fft, _ = dg01md('D', xr2.copy(), xi2.copy())
    xr_sum_fft, xi_sum_fft, _ = dg01md('D', xr_sum.copy(), xi_sum.copy())

    np.testing.assert_allclose(xr_sum_fft, a * xr1_fft + b * xr2_fft, rtol=1e-14)
    np.testing.assert_allclose(xi_sum_fft, a * xi1_fft + b * xi2_fft, rtol=1e-14)

def test_parseval_theorem():
    """
    Validate Parseval's theorem: sum|x|^2 = (1/N)*sum|X|^2.

    The energy in time domain equals energy in frequency domain (scaled).

    Random seed: 789 (for reproducibility)
    """
    np.random.seed(789)
    n = 16

    xr = np.random.randn(n)
    xi = np.random.randn(n)

    energy_time = np.sum(xr**2 + xi**2)

    xr_fft, xi_fft, info = dg01md('D', xr.copy(), xi.copy())
    assert info == 0

    energy_freq = np.sum(xr_fft**2 + xi_fft**2)

    np.testing.assert_allclose(energy_time, energy_freq / n, rtol=1e-14)

def test_dc_component():
    """
    Validate DC component (k=0) is sum of all samples.

    DFT[0] = sum(z[i]) for i=0..N-1

    Random seed: 111 (for reproducibility)
    """
    np.random.seed(111)
    n = 8

    xr = np.random.randn(n)
    xi = np.random.randn(n)

    xr_fft, xi_fft, info = dg01md('D', xr.copy(), xi.copy())
    assert info == 0

    np.testing.assert_allclose(xr_fft[0], np.sum(xr), rtol=1e-14)
    np.testing.assert_allclose(xi_fft[0], np.sum(xi), rtol=1e-14)

def test_real_signal_conjugate_symmetry():
    """
    Validate conjugate symmetry for real-valued input.

    For real input: X[k] = conj(X[N-k]) for k=1..N-1

    Random seed: 222 (for reproducibility)
    """
    np.random.seed(222)
    n = 8

    xr = np.random.randn(n)
    xi = np.zeros(n)

    xr_fft, xi_fft, info = dg01md('D', xr.copy(), xi.copy())
    assert info == 0

    for k in range(1, n // 2):
        np.testing.assert_allclose(xr_fft[k], xr_fft[n - k], rtol=1e-14)
        np.testing.assert_allclose(xi_fft[k], -xi_fft[n - k], rtol=1e-14)


"""Edge case tests."""

def test_n_equals_2():
    """
    Test minimum valid N=2.

    For N=2: DFT gives X[0]=x[0]+x[1], X[1]=x[0]-x[1]
    """
    xr = np.array([1.0, 2.0], dtype=float)
    xi = np.array([0.5, -0.5], dtype=float)

    xr_fft, xi_fft, info = dg01md('D', xr, xi)

    assert info == 0
    np.testing.assert_allclose(xr_fft[0], 3.0, rtol=1e-14)
    np.testing.assert_allclose(xi_fft[0], 0.0, rtol=1e-14)
    np.testing.assert_allclose(xr_fft[1], -1.0, rtol=1e-14)
    np.testing.assert_allclose(xi_fft[1], 1.0, rtol=1e-14)

def test_n_equals_4():
    """
    Test N=4.

    Random seed: 333 (for reproducibility)
    """
    np.random.seed(333)
    xr = np.random.randn(4)
    xi = np.random.randn(4)
    xr_orig = xr.copy()
    xi_orig = xi.copy()

    xr_fft, xi_fft, info1 = dg01md('D', xr, xi)
    assert info1 == 0

    xr_back, xi_back, info2 = dg01md('I', xr_fft, xi_fft)
    assert info2 == 0

    np.testing.assert_allclose(xr_back, 4 * xr_orig, rtol=1e-14)
    np.testing.assert_allclose(xi_back, 4 * xi_orig, rtol=1e-14)

def test_large_n_power_of_2():
    """
    Test larger N (128) to verify algorithm scales.

    Random seed: 444 (for reproducibility)
    """
    np.random.seed(444)
    n = 128
    xr = np.random.randn(n)
    xi = np.random.randn(n)
    xr_orig = xr.copy()
    xi_orig = xi.copy()

    xr_fft, xi_fft, info1 = dg01md('D', xr, xi)
    assert info1 == 0

    xr_back, xi_back, info2 = dg01md('I', xr_fft, xi_fft)
    assert info2 == 0

    np.testing.assert_allclose(xr_back, n * xr_orig, rtol=1e-12)
    np.testing.assert_allclose(xi_back, n * xi_orig, rtol=1e-12)

def test_zero_signal():
    """Test that zero signal gives zero output."""
    n = 8
    xr = np.zeros(n, dtype=float)
    xi = np.zeros(n, dtype=float)

    xr_fft, xi_fft, info = dg01md('D', xr, xi)

    assert info == 0
    np.testing.assert_allclose(xr_fft, np.zeros(n), atol=1e-15)
    np.testing.assert_allclose(xi_fft, np.zeros(n), atol=1e-15)

def test_impulse_response():
    """
    Test impulse at position 0 gives constant spectrum.

    delta[0]=1, delta[k]=0 for k>0 => DFT = [1, 1, 1, ..., 1]
    """
    n = 8
    xr = np.zeros(n, dtype=float)
    xi = np.zeros(n, dtype=float)
    xr[0] = 1.0

    xr_fft, xi_fft, info = dg01md('D', xr, xi)

    assert info == 0
    np.testing.assert_allclose(xr_fft, np.ones(n), rtol=1e-14)
    np.testing.assert_allclose(xi_fft, np.zeros(n), atol=1e-15)


"""Error handling tests."""

def test_invalid_indi():
    """Test with invalid INDI parameter."""
    xr = np.array([1.0, 2.0, 3.0, 4.0], dtype=float)
    xi = np.array([0.0, 0.0, 0.0, 0.0], dtype=float)
    xr_out, xi_out, info = dg01md('X', xr, xi)
    assert info == -1

def test_n_not_power_of_2():
    """Test with N not a power of 2."""
    xr = np.array([1.0, 2.0, 3.0], dtype=float)
    xi = np.array([0.0, 0.0, 0.0], dtype=float)
    xr_out, xi_out, info = dg01md('D', xr, xi)
    assert info == -2

def test_n_equals_1():
    """Test with N=1 (less than minimum 2)."""
    xr = np.array([1.0], dtype=float)
    xi = np.array([0.0], dtype=float)
    xr_out, xi_out, info = dg01md('D', xr, xi)
    assert info == -2

def test_lowercase_d():
    """Test that lowercase 'd' works for forward FFT."""
    xr = np.array([1.0, 2.0, 3.0, 4.0], dtype=float)
    xi = np.array([0.0, 0.0, 0.0, 0.0], dtype=float)
    xr_out, xi_out, info = dg01md('d', xr, xi)
    assert info == 0

def test_lowercase_i():
    """Test that lowercase 'i' works for inverse FFT."""
    xr = np.array([1.0, 2.0, 3.0, 4.0], dtype=float)
    xi = np.array([0.0, 0.0, 0.0, 0.0], dtype=float)
    xr_out, xi_out, info = dg01md('i', xr, xi)
    assert info == 0



"""
Tests for SB10TD - H2 optimal discrete-time controller transformation.

Transforms controller matrices from normalized system to original system.
"""
import numpy as np
import pytest
import slicot


class TestSB10TDBasic:
    """Basic functionality tests for SB10TD."""

    def test_basic_2x3x3(self):
        """
        Test basic transformation with a 2x3x3 system (n=2, m=3, np=3).

        Uses identity transformation matrices to verify pass-through behavior.
        Random seed: 42 (for reproducibility)
        """
        np.random.seed(42)

        n, m, np_ = 2, 3, 3
        ncon, nmeas = 1, 1

        d = np.zeros((np_, m), dtype=np.float64, order='F')
        d[np_ - nmeas:, m - ncon:] = 0.1

        tu = np.eye(ncon, dtype=np.float64, order='F')
        ty = np.eye(nmeas, dtype=np.float64, order='F')

        ak_in = np.array([[0.5, 0.1], [0.0, 0.4]], dtype=np.float64, order='F')
        bk_in = np.array([[0.2], [0.3]], dtype=np.float64, order='F')
        ck_in = np.array([[0.1, 0.2]], dtype=np.float64, order='F')
        dk_in = np.array([[0.05]], dtype=np.float64, order='F')

        ak, bk, ck, dk, rcond, info = slicot.sb10td(
            n, m, np_, ncon, nmeas,
            d, tu, ty, ak_in, bk_in, ck_in, dk_in
        )

        assert info == 0, f"SB10TD returned info = {info}"
        assert ak.shape == (n, n)
        assert bk.shape == (n, nmeas)
        assert ck.shape == (ncon, n)
        assert dk.shape == (ncon, nmeas)
        assert rcond > 0.0, "RCOND should be positive for well-conditioned problem"

    def test_transformation_formula(self):
        """
        Test that transformation follows expected formula.

        With identity TU, TY and zero D22:
        - BKHAT = BK * TY = BK
        - CKHAT = TU * CK = CK
        - DKHAT = TU * DK * TY = DK
        - Since D22 = 0, (I + DKHAT*D22) = I, so CK, DK unchanged
        - AK = AK_in - BK*D22*CK = AK_in when D22 = 0

        Random seed: 123 (for reproducibility)
        """
        np.random.seed(123)

        n, m, np_ = 3, 4, 4
        ncon, nmeas = 2, 2

        d = np.zeros((np_, m), dtype=np.float64, order='F')

        tu = np.eye(ncon, dtype=np.float64, order='F')
        ty = np.eye(nmeas, dtype=np.float64, order='F')

        ak_in = np.random.randn(n, n).astype(np.float64, order='F') * 0.1
        bk_in = np.random.randn(n, nmeas).astype(np.float64, order='F') * 0.1
        ck_in = np.random.randn(ncon, n).astype(np.float64, order='F') * 0.1
        dk_in = np.random.randn(ncon, nmeas).astype(np.float64, order='F') * 0.1

        ak, bk, ck, dk, rcond, info = slicot.sb10td(
            n, m, np_, ncon, nmeas,
            d, tu, ty,
            ak_in.copy(order='F'), bk_in.copy(order='F'),
            ck_in.copy(order='F'), dk_in.copy(order='F')
        )

        assert info == 0, f"SB10TD returned info = {info}"

        np.testing.assert_allclose(ak, ak_in, rtol=1e-14, atol=1e-15)
        np.testing.assert_allclose(bk, bk_in, rtol=1e-14, atol=1e-15)
        np.testing.assert_allclose(ck, ck_in, rtol=1e-14, atol=1e-15)
        np.testing.assert_allclose(dk, dk_in, rtol=1e-14, atol=1e-15)


class TestSB10TDNonTrivialTransform:
    """Test cases with non-trivial transformations."""

    def test_nontrivial_tu_ty(self):
        """
        Test with non-identity transformation matrices.

        Verifies correct application of TU and TY transformations.
        Random seed: 456 (for reproducibility)
        """
        np.random.seed(456)

        n, m, np_ = 2, 3, 3
        ncon, nmeas = 1, 1

        d = np.zeros((np_, m), dtype=np.float64, order='F')
        d[np_ - nmeas:, m - ncon:] = 0.0

        tu = np.array([[2.0]], dtype=np.float64, order='F')
        ty = np.array([[0.5]], dtype=np.float64, order='F')

        ak_in = np.array([[0.5, 0.1], [0.0, 0.4]], dtype=np.float64, order='F')
        bk_in = np.array([[0.2], [0.3]], dtype=np.float64, order='F')
        ck_in = np.array([[0.1, 0.2]], dtype=np.float64, order='F')
        dk_in = np.array([[0.1]], dtype=np.float64, order='F')

        ak, bk, ck, dk, rcond, info = slicot.sb10td(
            n, m, np_, ncon, nmeas,
            d, tu, ty,
            ak_in.copy(order='F'), bk_in.copy(order='F'),
            ck_in.copy(order='F'), dk_in.copy(order='F')
        )

        assert info == 0, f"SB10TD returned info = {info}"

        bk_expected = bk_in @ ty
        ck_expected = tu @ ck_in
        dk_expected = tu @ dk_in @ ty

        np.testing.assert_allclose(bk, bk_expected, rtol=1e-14, atol=1e-15)
        np.testing.assert_allclose(ck, ck_expected, rtol=1e-14, atol=1e-15)
        np.testing.assert_allclose(dk, dk_expected, rtol=1e-14, atol=1e-15)

    def test_with_nonzero_d22(self):
        """
        Test transformation with nonzero D22 submatrix.

        Verifies the full transformation when feedback loop exists.
        Random seed: 789 (for reproducibility)
        """
        np.random.seed(789)

        n, m, np_ = 3, 4, 4
        ncon, nmeas = 2, 2

        d = np.zeros((np_, m), dtype=np.float64, order='F')
        d22 = np.random.randn(nmeas, ncon).astype(np.float64) * 0.1
        d[np_ - nmeas:, m - ncon:] = d22

        tu = np.eye(ncon, dtype=np.float64, order='F')
        ty = np.eye(nmeas, dtype=np.float64, order='F')

        ak_in = np.random.randn(n, n).astype(np.float64, order='F') * 0.2
        bk_in = np.random.randn(n, nmeas).astype(np.float64, order='F') * 0.2
        ck_in = np.random.randn(ncon, n).astype(np.float64, order='F') * 0.2
        dk_in = np.random.randn(ncon, nmeas).astype(np.float64, order='F') * 0.2

        ak, bk, ck, dk, rcond, info = slicot.sb10td(
            n, m, np_, ncon, nmeas,
            d, tu, ty,
            ak_in.copy(order='F'), bk_in.copy(order='F'),
            ck_in.copy(order='F'), dk_in.copy(order='F')
        )

        assert info == 0, f"SB10TD returned info = {info}"
        assert ak.shape == (n, n)
        assert bk.shape == (n, nmeas)
        assert ck.shape == (ncon, n)
        assert dk.shape == (ncon, nmeas)
        assert rcond > 0.0

        bkhat = bk_in @ ty
        ckhat = tu @ ck_in
        dkhat = tu @ dk_in @ ty

        Im2_plus_dkhat_d22 = np.eye(ncon) + dkhat @ d22
        ck_expected = np.linalg.solve(Im2_plus_dkhat_d22, ckhat)
        dk_expected = np.linalg.solve(Im2_plus_dkhat_d22, dkhat)

        temp = bkhat @ d22
        ak_expected = ak_in - temp @ ck_expected
        bk_expected = bkhat - temp @ dk_expected

        np.testing.assert_allclose(ak, ak_expected, rtol=1e-13, atol=1e-14)
        np.testing.assert_allclose(bk, bk_expected, rtol=1e-13, atol=1e-14)
        np.testing.assert_allclose(ck, ck_expected, rtol=1e-13, atol=1e-14)
        np.testing.assert_allclose(dk, dk_expected, rtol=1e-13, atol=1e-14)


class TestSB10TDEdgeCases:
    """Edge case tests for SB10TD."""

    def test_zero_order_system(self):
        """Test with n=0 (zero-order system)."""
        n, m, np_ = 0, 2, 2
        ncon, nmeas = 1, 1

        d = np.zeros((np_, m), dtype=np.float64, order='F')
        tu = np.eye(ncon, dtype=np.float64, order='F')
        ty = np.eye(nmeas, dtype=np.float64, order='F')

        ak_in = np.zeros((1, 0), dtype=np.float64, order='F').reshape(0, 0, order='F')
        bk_in = np.zeros((0, nmeas), dtype=np.float64, order='F')
        ck_in = np.zeros((ncon, 0), dtype=np.float64, order='F')
        dk_in = np.array([[0.1]], dtype=np.float64, order='F')

        ak, bk, ck, dk, rcond, info = slicot.sb10td(
            n, m, np_, ncon, nmeas,
            d, tu, ty, ak_in, bk_in, ck_in, dk_in
        )

        assert info == 0

    def test_single_control_single_measurement(self):
        """
        Test SISO controller (ncon=1, nmeas=1).

        Random seed: 999 (for reproducibility)
        """
        np.random.seed(999)

        n, m, np_ = 4, 3, 3
        ncon, nmeas = 1, 1

        d = np.zeros((np_, m), dtype=np.float64, order='F')
        d[np_ - nmeas:, m - ncon:] = 0.05

        tu = np.array([[1.5]], dtype=np.float64, order='F')
        ty = np.array([[0.8]], dtype=np.float64, order='F')

        ak_in = np.random.randn(n, n).astype(np.float64, order='F') * 0.1
        bk_in = np.random.randn(n, nmeas).astype(np.float64, order='F') * 0.1
        ck_in = np.random.randn(ncon, n).astype(np.float64, order='F') * 0.1
        dk_in = np.random.randn(ncon, nmeas).astype(np.float64, order='F') * 0.1

        ak, bk, ck, dk, rcond, info = slicot.sb10td(
            n, m, np_, ncon, nmeas,
            d, tu, ty,
            ak_in.copy(order='F'), bk_in.copy(order='F'),
            ck_in.copy(order='F'), dk_in.copy(order='F')
        )

        assert info == 0
        assert ak.shape == (n, n)
        assert bk.shape == (n, nmeas)
        assert ck.shape == (ncon, n)
        assert dk.shape == (ncon, nmeas)


class TestSB10TDErrorHandling:
    """Error handling tests for SB10TD."""

    def test_singular_matrix_error(self):
        """
        Test that info=1 is returned when (I + DKHAT*D22) is singular.
        """
        n, m, np_ = 2, 2, 2
        ncon, nmeas = 1, 1

        d = np.zeros((np_, m), dtype=np.float64, order='F')
        d[np_ - nmeas:, m - ncon:] = 10.0

        tu = np.eye(ncon, dtype=np.float64, order='F')
        ty = np.eye(nmeas, dtype=np.float64, order='F')

        ak_in = np.array([[0.5, 0.1], [0.0, 0.4]], dtype=np.float64, order='F')
        bk_in = np.array([[0.2], [0.3]], dtype=np.float64, order='F')
        ck_in = np.array([[0.1, 0.2]], dtype=np.float64, order='F')
        dk_in = np.array([[-0.1]], dtype=np.float64, order='F')

        ak, bk, ck, dk, rcond, info = slicot.sb10td(
            n, m, np_, ncon, nmeas,
            d, tu, ty,
            ak_in.copy(order='F'), bk_in.copy(order='F'),
            ck_in.copy(order='F'), dk_in.copy(order='F')
        )

        assert info == 1, f"Expected info=1 for singular matrix, got {info}"
